#include "core/core.h"
#include <windows.h>
#include <winsock2.h>
#include <process.h>

#undef RGB

namespace bpp
{

#include "System.h"
#include "Sockets.h"

namespace Sockets
{

using namespace System;

static __stdcall unsigned int SocketListener(void* ptr)
{
	Socket& s = *(Socket*)ptr;
	fd_set fd;
	FD_ZERO(&fd);
	FD_SET((SOCKET)s.Handle(), &fd);
	if (select(0, &fd, NULL, NULL, NULL) > 0)
	{
		GenericEvent *evt = new GenericEvent;
		evt->Sender = &s;
		s.OnReceive.raise(bpp::ref<GenericEvent>(evt));
	}
	return 0;
}


void Socket::Initialize()
{
	Create();
	mListener = _beginthreadex(NULL, 0, SocketListener, this,
		CREATE_SUSPENDED, (unsigned int*)&mListener);
}

	
void Socket::Terminate()
{
	TerminateThread((HANDLE)mListener, 0);
	shutdown(mHandle, SD_BOTH);
	closesocket(mHandle);
}


void Socket::Create()
{
	mHandle = socket(PF_INET, SOCK_STREAM, IPPROTO_IP);
}


void Socket::Disconnect()
{
	DisableListener();
	shutdown(mHandle, SD_BOTH);
	closesocket(mHandle);
	Create();
}


int Socket::ReadBlock(void* buf, int size)
{
	int result = recv(mHandle, (char*)buf, size, 0);
	return result == SOCKET_ERROR ? 0 : result;
}


int Socket::WriteBlock(void* buf, int size)
{
	int result = send(mHandle, (char*)buf, size, 0);
	return result == SOCKET_ERROR ? 0 : result;
}


void Socket::EnableListener()
{
	ResumeThread((HANDLE)mListener);
}


void Socket::DisableListener()
{
	SuspendThread((HANDLE)mListener);
}


bool Socket::Connected()
{
	unsigned long nb = true;
	ioctlsocket(mHandle, FIONBIO, &nb);
	char buf;
	int result = recv(mHandle, &buf, 1, MSG_PEEK);
	int err = WSAGetLastError();
	nb = false;
	ioctlsocket(mHandle, FIONBIO, &nb);
	if (result == SOCKET_ERROR)
		return (err == WSAEWOULDBLOCK);
	else
		return result;
}


bool Socket::DataReceived()
{
	if (!Connected())
		return false;
	timeval tv = { 0, 0 };
	fd_set fd;
	FD_ZERO(&fd);
	FD_SET((SOCKET)mHandle, &fd);
	return (select(0, &fd, NULL, NULL, &tv) > 0);
}


int Socket::Handle()
{
	return mHandle;
}


void Socket::Handle(int value)
{
	DisableListener();
	closesocket(mHandle);
	mHandle = value;
	EnableListener();
}


int ClientSocket::ConnectByIP(string ip, unsigned short port)
{
	Disconnect();
	sockaddr_in sa;
	sa.sin_family = PF_INET;
	sa.sin_addr.s_addr = inet_addr(ip.cstr());
	sa.sin_port = htons(port);
	bool result = !connect(mHandle, (sockaddr*)&sa, sizeof(sa));
	return result;
}


bool ClientSocket::ConnectByName(string hostname, unsigned short port)
{
	Disconnect();
	hostent* host = gethostbyname(hostname.cstr());
	if (!host)
		return NULL;
	sockaddr_in sa;
	sa.sin_family = PF_INET; 
	sa.sin_addr.s_addr = *((u_long FAR *) (host->h_addr));
	sa.sin_port = htons(port); 
	bool result = !connect(mHandle, (sockaddr*)&sa, sizeof(sa));
	return result;
}


bool ClientSocket::Connect(string host, unsigned short port)
{
	Disconnect();
	sockaddr_in sa;
	sa.sin_family = PF_INET; 
	sa.sin_addr.s_addr = inet_addr(host.cstr());
	if (sa.sin_addr.s_addr == INADDR_NONE)
	{
		hostent* hst = gethostbyname(host.cstr());
		sa.sin_addr.s_addr = *((u_long FAR *) (hst->h_addr));
	}
	sa.sin_port = htons(port); 
	bool result = !connect(mHandle, (sockaddr*)&sa, sizeof(sa));
	return result;
}


static __stdcall unsigned int ServerSocketListener(void* ptr)
{
	ServerSocket& s = *(ServerSocket*)ptr;
	while (true)
	{
		int ah = accept(s.Handle(), NULL, NULL);
		if (ah != INVALID_SOCKET)
		{
			bpp::ref<Socket> as = new Socket;
			as->Handle(ah);
			s.Connect(as);
		}
	}
	return 0;
}


void ServerSocket::Connect(bpp::ref<Socket> s)
{
	ConnectionEvent *evt = new ConnectionEvent;
	evt->Sender = this;
	evt->Socket = s;
	OnConnect.raise(bpp::ref<ConnectionEvent>(evt));
}


void ServerSocket::Initialize()
{
	mHandle = socket(PF_INET, SOCK_STREAM, IPPROTO_IP);
	mListener = _beginthreadex(NULL, 0, ServerSocketListener, this,
		CREATE_SUSPENDED, (unsigned int*)&mListener);
}


void ServerSocket::Terminate()
{
	TerminateThread((HANDLE)mListener, 0);
	shutdown(mHandle, SD_BOTH);
	closesocket(mHandle);
}


bpp::ref<Socket> ServerSocket::Accept()
{
	int ah = accept(mHandle, NULL, NULL);
	if (ah != INVALID_SOCKET)
	{
		bpp::ref<Socket> as = new Socket;
		as->Handle(ah);
		return as;
	}
	else
		return 0;
}


bool ServerSocket::PendingConnection()
{
	timeval tv = { 0, 0 };
	fd_set fd;
	FD_ZERO(&fd);
	FD_SET(mHandle, &fd);
	return (select(0, &fd, NULL, NULL, &tv) > 0);
}


unsigned short ServerSocket::Port()
{
	return mPort;
}


int ServerSocket::Handle()
{
	return mHandle;
}


bool ServerSocket::Background()
{
	return mBackground;
}


void ServerSocket::Background(bool value)
{
	if (value)
		ResumeThread((HANDLE)mListener);
	else
		SuspendThread((HANDLE)mListener);
	mBackground = value;
}


void ServerSocket::Port(unsigned short value)
{
	sockaddr_in sa;
	sa.sin_family = PF_INET; 
	sa.sin_addr.s_addr = 0;
	sa.sin_port = htons(value);
	bind(mHandle, (sockaddr*)&sa, sizeof(sa));
	listen(mHandle, SOMAXCONN);
	mPort = value;
}


Sockets::initialization::initialization()
{
	WSADATA wsadata;
	WSAStartup(MAKEWORD(2, 0), &wsadata);
}


Sockets::finalization::~finalization()
{
	WSACleanup();
}

}
}