#include "ident.h"
#include "module.h"
#include "util.h"
#include <algorithm>


ident::ident(FILE* in, module* mod)
{
	name = freads(in);
	scope = fgetc(in);
	owner = mod;
}


ident::ident(int kind, const string& name, int scope)
{
	this->kind = kind;
	this->name = name;
	this->scope = scope;
	owner = NULL;
}


string ident::deftype() const
{
	switch (name[name.length()-1])
	{
	case '%':
		return "SHORT";
	case '&':
		return "LONG";
	case '!':
		return "SINGLE";
	case '#':
		return "DOUBLE";
	case '@':
		return "DECIMAL";
	case '$':
		return "STRING";
	default:
		return "VARIANT";
	}
}


string ident::cname(bool full) const
{
	string suffix = deftype();
	string result;
	if (full && owner && scope != local)
	{
		result = owner->name + "::";
	}
	if (suffix.size() && !(suffix ^ "VARIANT"))
		result += name.substr(0, name.length() - 1) + "_" + suffix;
	else
		result += name;
	return result;
}


void ident::save(FILE* out) const
{
	fputc(kind, out);
	fwrites(name, out);
	fputc(scope, out);
}


ident* ident::load(FILE* in, module* mod)
{
	ident* id;
	int kind = fgetc(in);
	switch (kind)
	{
	case cons:
		id = new constant(in, mod); break;
	case type:
		id = new udt(in, mod); break;
	case var:
		id = new variable(in, mod); break;
	case func:
		id = new function(in, mod); break;
	case struc:
		id = new structure(in, mod); break;
	case cls:
		id = new clas(in, mod); break;
	case als:
		id = new alias(in, mod); break;
	default:
		id = NULL;
	}
	if (id)
		id->kind = kind;
	return id;
}


void ident::write(ostream& out) const
{
}


alias::alias(FILE* in, module* mod):
	ident(in, mod)
{
	type = fgetc(in);
	if (type)
	{
		aliased = mod->types[freads(in)];
		mod->types.push_back((udt*)this);
	}
	else
		aliased = mod->idents[freads(in)];
}


void alias::save(FILE* out) const
{
	ident::save(out);
	fputc(type, out);
	fwrites(aliased->name, out);
}


constant::constant(FILE* in, module* mod):
	ident(in, mod)
{
	str = fgetc(in);
	value = freads(in);
}


constant::constant(const string& name, int scope):
	ident(cons, name, scope)
{
}


void constant::save(FILE* out) const
{
	ident::save(out);
	fputc(str, out);
	fwrites(value, out);
}


void constant::write(ostream& out) const
{
	out << "const ";
	if (str)
		out << "string ";
	else
	{
		double n = atof(value.c_str());
		if (n - abs(n))
			out << "double ";
	}
	out << cname(false) << " = " << value << ";\n";
}


udt::udt(FILE* in, module* mod):
	ident(in, mod)
{
	mod->types.push_back(this);
}


udt::udt(const string& name):
	ident(type, name, pub)
{
}


udt::udt(int kind, const string& name, int scope):
	ident(kind, name, scope)
{
}


void udt::put(FILE* out) const
{
	if (owner)
		fwrites(owner->name, out);
	else
		fputc(0, out);
	fwrites(name, out);
}


string udt::cname(bool full) const
{
	if (name ^ "BYTE")
		return "unsigned char";
	else if (name ^ "WORD")
		return "unsigned short";
	else if (name ^ "DWORD")
		return "unsigned long";
	else if (name ^ "SHORT")
		return "short";
	else if (name ^ "LONG")
		return "int";
	else if (name ^ "INTEGER")
		return "int";
	else if (name ^ "SINGLE")
		return "float";
	else if (name ^ "DOUBLE")
		return "double";
	else if (name ^ "DECIMAL")
		return "long double";
	else if (name ^ "BOOLEAN")
		return "bool";
	else if (name ^ "STRING")
		return "string";
	else if (name ^ "APISTRING")
		return "apistring";
	else if (name ^ "VARIANT")
		return "variant";
	else if (name ^ "EVENT")
		return "event";
	else if (name ^ "ANY")
		return "void*";
	else if (name ^ "OBJECT")
		return "object";
	else
		return ident::cname(full);
}


variable::variable(FILE* in, module* mod):
	ident(in, mod)
{
	type = mod->gettype(in);
	array = fgetc(in);
	byref = fgetc(in);
	propget = fgetc(in);
	propset = fgetc(in);
	col = fgetc(in);
	flstr = fgeti(in);
	statar = fgeti(in);
	for (int i = 0; i < statar; i++)
		stdims[i] = fgeti(in);
	defval = freads(in);
}


variable::variable(const string& name, int scope, int level, udt* type):
	ident(var, name, scope)
{
	this->level = level;
	this->type = type;
	byref = array = propget = propset = false;
	flstr = col = statar = 0;
}


void variable::save(FILE* out) const
{
	ident::save(out);
	type->put(out);
	fputc(array, out);
	fputc(byref, out);
	fputc(propget, out);
	fputc(propset, out);
	fputc(col, out);
	fputi(flstr, out);
	fputi(statar, out);
	for (int i = 0; i < statar; i++)
		fputi(stdims[i], out);
	fwrites(defval, out);
}


void variable::write(ostream& out) const
{
	out << "extern ";
	if (array)
		out << "bpp::array<";
	if (col == variable::coll)
		out << "bpp::collection<";
	else if (col == variable::dict)
		out << "bpp::dictionary<";
	if (type->kind == udt::cls)
		out << "bpp::ref<class ";
	out << type->cname();
	if (array)
		out << "> ";
	if (col)
		out << "> ";
	if (type->kind == udt::cls)
		out << ">";
	out << " " << cname(false) << ";\n";
}


function::function(FILE* in, module* mod):
	ident(in, mod)
{
	if (fgetc(in))
		type = owner->gettype(in);
	else
		type = NULL;
	method = fgetc(in);
	external = fgetc(in);
	generic = fgetc(in);
	int argcnt = fgeti(in);
	args.reserve(argcnt);
	for (int i = 0; i < argcnt; i++)
		args.push_back((variable*)load(in, mod));
}


function::function(const string& name, int scope, bool resolved,
	bool method, udt* type):
	ident(func, name, scope)
{
	this->type = type;
	this->method = method;
	this->resolved = resolved;
	external = generic = false;
}


void function::save(FILE* out) const
{
	ident::save(out);
	if (type)
	{
		fputc(true, out);
		type->put(out);
	}
	else
		fputc(false, out);
	fputc(method, out);
	fputc(external, out);
	fputc(generic, out);
	fputi(args.size(), out);
	table<variable*>::const_iterator i;
	for (i = args.begin(); i != args.end(); i++)
		(*i)->save(out);
}


void function::write(ostream& out) const
{
	if (generic)
	{
/*
		out << "virtual void " << cname(false) << "(unsigned char*);\n"
			<< "virtual void " << cname(false) << "(unsigned short*);\n"
			<< "virtual void " << cname(false) << "(unsigned long*);\n"
			<< "virtual void " << cname(false) << "(short*);\n"
			<< "virtual void " << cname(false) << "(int*);\n"
			<< "virtual void " << cname(false) << "(float*);\n"
			<< "virtual void " << cname(false) << "(double*);\n"
			<< "virtual void " << cname(false) << "(long double*);\n"
			<< "virtual void " << cname(false) << "(string*);\n"
			<< "virtual void " << cname(false) << "(variant*);\n"
			<< "virtual void " << cname(false) << "(void*);\n";
*/
		out << "\t\t\tvirtual void " << cname(false)
			<< "(bpp::string& x);\n";
		out << "\t\t\tvirtual void " << cname(false)
			<< "(bpp::apistring& x);\n";
		out << "\t\t\tvirtual void " << cname(false)
			<< "(ref<bpp::object>& x);\n";
//		out << "\t\t\ttemplate<int n> void " << cname(false)
//			<< "(bpp::flstring<n>& x) { " << cname(false) << "(&x, n); }\n";
		out << "\t\t\ttemplate<class T> void " << cname(false)
			<< "(bpp::ref<T>& x) { bpp::ref<bpp::object> y = x; " << cname(false) << "(y); }\n";
		out << "\t\t\ttemplate<class T> void " << cname(false)
			<< "(T& x) { " << cname(false) << "(&x, sizeof(T)); }\n";
		out << "\t\t\tvirtual void " << cname(false)
			<< "(void*, int);\n";
		return;
	}
	if (external)
		out << "extern \"C\" ";
	if (type)
	{
		if (type->kind == udt::cls)
			out << "bpp::ref<class ";
		out << type->cname();
	}
	else
		out << "void";
	if (type)
		if (type->kind == udt::cls)
			out << ">";
	out << " ";
	if (!method)
		out << "__stdcall ";
	out << cname(false) << "(";
	table<variable*>::const_iterator i;
	for (i = args.begin(); i != args.end(); i++)
	{
		if (i != args.begin())
			out << ", ";
		if ((*i)->array)
			out << "bpp::array<";
		if ((*i)->col == variable::coll)
			out << "bpp::collection<";
		else if ((*i)->col == variable::dict)
			out << "bpp::dictionary<";
		if ((*i)->type->kind == udt::cls)
			out << "bpp::ref<class ";
		out << (*i)->type->cname();
		if ((*i)->byref)
			out << "&";
		if ((*i)->col)
			out << "> ";
		if ((*i)->array)
			out << "> ";
		if ((*i)->type->kind == udt::cls)
			out << ">";
	}
	out << ");\n";
}


structure::structure(FILE* in, module* mod):
	udt(in, mod)
{
	int fldcnt = fgeti(in);
	fields.reserve(fldcnt);
	for (int i = 0; i < fldcnt; i++)
		fields.push_back((variable*)load(in, mod));
}


structure::structure(const string& name, int scope):
	udt(struc, name, scope)
{
}


void structure::save(FILE* out) const
{
	ident::save(out);
	table<variable*>::const_iterator i;
	fputi(fields.size(), out);
	for (i = fields.begin(); i != fields.end(); i++)
		(*i)->save(out);
}


void structure::write(ostream& out) const
{
	if (name ^ "EVENT" || name ^ "ANY")
		return;
	out << "\nstruct " << cname(false) << "\n{";
	table<variable*>::const_iterator i;
	for (i = fields.begin(); i != fields.end(); i++)
	{
		out << "\n\t";
		if ((*i)->type->kind == udt::cls)
			out << "bpp::ref<class ";
		if ((*i)->flstr)
			out << "bpp::flstring<" << (*i)->flstr << "> ";
		else
			out << (*i)->type->cname();
		if ((*i)->type->kind == udt::cls)
			out << ">";
		out << " " << (*i)->cname(false);
		if ((*i)->statar)
		{
			for (int j = 0; j < 5; j++)
			{
				if (!(*i)->stdims[j])
					break;
				out << "[" << (*i)->stdims[j] << "]";
			}
		}
		out << ";";
	}
	out << "\n};\n\n";
}


clas::clas(FILE* in, module* mod):
	udt(in, mod)
{
	if (fgetc(in))
		super = (clas*)owner->gettype(in);
	else
		super = NULL;
	{
		int propcnt = fgeti(in);
		props.reserve(propcnt);
		for (int i = 0; i < propcnt; i++)
			props.push_back((variable*)load(in, mod));
	}
	{
		int mtdcnt = fgeti(in);
		methods.reserve(mtdcnt);
		for (int i = 0; i < mtdcnt; i++)
			methods.push_back((function*)load(in, mod));
	}
	resolved = true;
}

	
clas::clas(const string& name, int scope, clas* super):
	udt(cls, name, scope)
{
	this->super = super;
	resolved = true;
}


variable* clas::prop(const string& name)
{
	try
	{
		return props[name];
	}
	catch (out_of_range)
	{
		if (super)
			return super->prop(name);
		else
			throw;
	}
}


function* clas::method(const string& name)
{
	try
	{
		return methods[name];
	}
	catch (out_of_range)
	{
		if (super)
			return super->method(name);
		else
			throw;
	}
}


ident* clas::member(const string& name, bool inherited)
{
	if (inherited)
	{
		try
		{
			return prop(name);
		}
		catch (out_of_range)
		{
			return method(name);
		}
	}
	else
	{
		try
		{
			return props[name];
		}
		catch (out_of_range)
		{
			return methods[name];
		}
	}
}


bool clas::childof(clas* c)
{
	if (!this)
		return false;
	clas* sup = this;
	while (sup != c)
		if (sup)
			sup = sup->super;
		else
			return false;
	return true;
}


void clas::save(FILE* out) const
{
	ident::save(out);
	if (super)
	{
		fputc(true, out);
		super->put(out);
	}
	else
		fputc(false, out);
	{
		fputi(props.size(), out);
		table<variable*>::const_iterator i;
		for (i = props.begin(); i != props.end(); i++)
			(*i)->save(out);
	}
	{
		fputi(methods.size(), out);
		table<function*>::const_iterator i;
		for (i = methods.begin(); i != methods.end(); i++)
			(*i)->save(out);
	}
}


void clas::write(ostream& out) const
{
	if (name ^ "OBJECT")
		return;
	out << "\nclass " << cname(false);
	if (super)
		out << ": public " << super->cname();
	out << "\n{\n";
	out << "public: " << cname(false) << "(bool init = true): "
		<< super->cname() << "(false) { if (init) { __refcnt++; Initialize(); __refcnt--; } }\n";
	out << "virtual bool __isbase(const type_info& ti) "
		<< "{ return (typeid(" << cname() << ") == ti) || "
		<< super->cname() << "::__isbase(ti); }\n";
	out << "virtual bpp::string ClassName() "
		<< "{ return \"" << owner->name << "." << name << "\"; }\n";
	{
		table<variable*>::const_iterator i;
		for (i = props.begin(); i != props.end(); i++)
		{
			if ((*i)->propget || (*i)->propset)
			{
				out << "\n";
				if ((*i)->propget)
				{
					out << "\t\t\tvirtual ";
					if ((*i)->type->kind == udt::cls)
						out << "bpp::ref<class ";
					out << (*i)->type->cname();
					if ((*i)->type->kind == udt::cls)
						out << ">";
					out << " " << (*i)->cname(false) << "();\n";
				}
				if ((*i)->propset)
				{
					out << "\t\t\tvirtual void " << (*i)->cname(false) << "(";
					if ((*i)->type->kind == udt::cls)
						out << "bpp::ref<class ";
					out << (*i)->type->cname();
					if ((*i)->type->kind == udt::cls)
						out << ">";
					out << ");\n";
				}
			}
			else
			{
				if ((*i)->array)
					out << "bpp::array<";
				if ((*i)->col == variable::coll)
					out << "bpp::collection<";
				else if ((*i)->col == variable::dict)
					out << "bpp::dictionary<";
				if ((*i)->type->kind == udt::cls)
					out << "bpp::ref<class ";
				out << (*i)->type->cname();
				if ((*i)->col)
					out << "> ";
				if ((*i)->array)
					out << "> ";
				if ((*i)->type->kind == udt::cls)
					out << ">";
				out << " " << (*i)->cname(false) << ";\n";
			}
		}
	}
	{
		table<function*>::const_iterator i;
		for (i = methods.begin(); i != methods.end(); i++)
		{
			if ((*i)->name ^ "ClassName")
				continue;
			if (!(*i)->generic)
				out << "virtual ";
			(*i)->write(out);
		}
	}
	out << "};\n";
}
