#include "parser.h"
#include "util.h"
#include <sstream>
#include <set>


parser::parser(string name, const string& basepath):
	in(addfileext(proper(name), "bpp").c_str(), ios::in | ios::binary),
	lex(in),
	me(proper(name))
{
	name = proper(name);
	if (in.fail())
 		throw "Error opening input file";
	if (out.fail())
		throw "Error opening output file";
	if (reserved(name))
		throw "Error: illegal module name '" + name + "'";
	this->basepath = basepath;
	modules.push_back(&me);
	me.idents.reserve(100000);
	me.types.reserve(10000);
	registerbuiltins();
	indent = level = loops = 0;
	inside = NULL;
	console = true;
	vbdll = explic = dll = apistr = devmode = false;
	resfile = "";
}


parser::~parser()
{
	for (table<module*>::iterator i = modules.begin() + 1; i != modules.end(); i++)
		delete *i;
}


void parser::registerbuiltins()
{
	function* Main = new function("Main", ident::pub, false);
	Main->external = true;
	me.idents.push_back(Main);
}


string parser::newline()
{
	return "\n" + string(indent, '\t');
}


udt* parser::findtype(string udtname, module* mod)
{
	if (apistr && udtname ^ "STRING")
		udtname = "APISTRING";
	if (mod)
		return mod->types[udtname];
	else
	{
		table<module*>::iterator i;
		for (i = modules.begin(); i != modules.end(); i++)
		{
			try
			{
				return (*i)->types[udtname];
			}
			catch (out_of_range)
			{
			}
		}
	}
	throw out_of_range("type not found");
}


ident* parser::findid(string idname, module* mod)
{
	if (mod)
		return mod->idents[idname];
	else
	{
		try
		{
			return me.idents[idname];
		}
		catch (out_of_range)
		{
			table<module*>::iterator i = modules.end() - 1;
			do
			{
				try
				{
					return (*i)->idents[idname];
				}
				catch (out_of_range) { }
				i--;
			} while (i != modules.begin());
		}
	}
	throw out_of_range("identifier not found");
}


void parser::enter()
{
	me.idents.push_back(NULL);
	level++;
}


void parser::leave()
{
	while (me.idents.back())
		me.idents.pop_back();
	me.idents.pop_back();
	level--;
}


void parser::parse(bool lib, bool devmode)
{
	parse_error err;
	this->lib = lib;
	this->devmode = devmode;
	try
	{
		parse_module();
	}
	catch (const char* message)
	{
		err.error = message;
		err.module = me.name;
		err.line = lex.line;
		err.col = lex.col;
		throw err;
	}
	catch (string message)
	{
		err.error = message;
		err.module = me.name;
		err.line = lex.line;
		err.col = lex.col;
		throw err;
	}
	if (lib)
		me.save(*this);
}


void parser::compile(bool verbose)
{
	out.close();
	if (inter)
		return;
	system(("bcc32 -d -O2 -c -w- -WM " + me.name + ".cpp").c_str());
	if (lib)
	{
		system(("tlib /P32 " + me.name + ".lib -+" +
			me.name + (verbose ? "" : "")).c_str());
		remove((me.name + ".bak").c_str());
	}
	else
	{
		string cmdline = "ilink32 -Gn ";
		if (dll)
			cmdline += "-aa -Tpd c0d32.obj ";
		else if (console)
			cmdline += "-ap c0x32.obj " + basepath +
				(devmode ? "lib\\core\\stc_dev.obj " : "lib\\core\\stc.obj ");
		else
			cmdline += "-aa -Tpe c0w32.obj " + basepath +
				(devmode ? "lib\\core\\stw_dev.obj " : "lib\\core\\stw.obj ");
		cmdline += basepath + 
			(devmode ? "lib\\core\\core_dev.obj " : "lib\\core\\core.obj ") +
			me.name + ".obj ";
		cmdline += "," + me.name + (dll ? ".dll" : ".exe") + ",nul,cw32mt.lib import32.lib";
		table<module*>::iterator i;
		for (i = modules.begin() + 1; i != modules.end(); i++)
			cmdline += " " + (*i)->path + (*i)->name + ".lib";
		if (!resfile.empty())
			cmdline += ",," + changefileext(resfile, "res");
		system(cmdline.c_str());
		remove((me.name + ".cpp").c_str());
		remove((me.name + ".tds").c_str());
	}
	remove((me.name + ".obj").c_str());
	vector<string>::iterator i;
	for (i = trash.begin(); i != trash.end(); i++)
		remove(i->c_str());
}


void parser::write()
{
	me.write(*this, "");
}


bool parser::reserved(const string& token)
{
	static const char* keywords[] = {
		"PUBLIC", "PRIVATE", "PROTECTED", "PUBLISHED", "DIM", "SUB",
		"FUNCTION", "PROPERTY", "SET", "GET", "LET", "END", "GOTO"
		"CONST", "AS", "USING", "INTERFACE", "OPTION",
		"EXIT", "LET", "CALL", "IF", "THEN", "ELSE",
		"ELSEIF", "DO", "WHILE", "UNTIL", "LOOP", "WEND",
		"FOR", "TO", "STEP", "NEXT", "SELECT", "CASE",
		"TRY", "CATCH", "THROW", "PRINT", "INPUT", "BYTE",
		"SHORT", "LONG", "INTEGER", "WORD", "DWORD", "ENUM",
		"SINGLE", "DOUBLE", "DECIMAL", "BOOLEAN", "OBJECT",
		"VARIANT", "CLASS", "EXTENDS", "TYPE", "ALIAS",
		"BUILTIN", "DECLARE", "EXTERNAL", "INITIALIZATION",
		"FINALIZATION", "COLLECTION", "DICTIONARY", "PRINT",
		"INPUT", "ITERATE", "INCR", "DECR", "SWAP", "EXPORT", 0 };
	for (const char** p = keywords; *p; p++)
		if (token ^ *p)
			return true;
	return false;
}


void parser::validate(const string& id)
{
	if (reserved(id))
		throw "illegal use of reserved word '" + id +
			"' as identifier";
}


void parser::parse_module()
{
	lex.next();
	inter = false;
	if (lex.token ^ "INTERFACE")
	{
		out.open("nul", ios::out | ios::binary);
		lex.next();
		if (lex.token ^ "NOSKIP")
			lex.next();
		else
			inter = true;
		if (lex.tokkind != lexer::nl)
			throw "end of statement expected";
		lex.next();
	}
	else
	{
		out.open(changefileext(me.name, "cpp").c_str(), ios::out | ios::binary);
		if (devmode)
			out << "#define __BPP_DEBUG\n";
		out << "#include \"" << basepath << "lib/core/core.h\"\n"
			<< "\nnamespace bpp\n{\nnamespace " << me.name << "\n{\n\n";
	}
	if (!(me.name ^ "SYSTEM"))
		modules.push_back(new module("System", *this, lib));
	out << newline();
	while (true)
	{
		while (lex.tokkind == lexer::nl && !in.eof())
		{
			out << newline();
			lex.next();
		}
		if (in.eof())
			break;
		int scope = ident::priv;
		if (lex.token ^ "PRIVATE")
			lex.next();
		else if (lex.token ^ "PUBLIC")
		{
			scope = ident::pub;
			lex.next();
		}
		if (lex.token ^ "OPTION")
			parse_option();
		else if (lex.token ^ "CONST")
			parse_const(scope);
		else if (lex.token ^ "ENUM")
			parse_enum(scope);
		else if (lex.token ^ "INITIALIZATION")
			parse_initialization();
		else if (lex.token ^ "FINALIZATION")
			parse_finalization();
		else if (lex.token ^ "DECLARE")
		{
			lex.next();
			if (lex.token ^ "BUILTIN")
				parse_builtin(scope);
			else if (lex.token ^ "INITIALIZATION")
			{
				me.init = true;
				lex.next();
			}
			else if (lex.token ^ "FINALIZATION")
			{
				me.final = true;
				lex.next();
			}
			else if (lex.token ^ "CLASS")
				parse_classdecl(scope);
			else
				parse_func(scope);
		}
		else if (lex.token ^ "SUB" || lex.token ^ "FUNCTION" || lex.token ^ "EXPORT")
			parse_func(scope, false);
		else if (lex.token ^ "PROPERTY")
			parse_propgs(scope);
		else if (lex.token ^ "DIM")
			parse_dim(scope);
		else if (lex.token ^ "TYPE")
			parse_type(scope);
		else if (lex.token ^ "CLASS")
			parse_class(scope);
		else if (lex.token ^ "USING")
			parse_using();
		else if (lex.token ^ "ALIAS")
			parse_alias();
		else if (lex.tokkind != lexer::nl)
			parse_dim(scope);
		if (lex.tokkind != lexer::nl)
			throw "end of statement expected";
		out << newline();
		lex.next();
	}
	if (!me.init)
	{
		out << "\nstruct initialization { initialization(); };" << newline();
		out << "initialization::initialization()";
		write_dll_loader();
		out << newline();
		me.init = true;
	}
	if (!me.final)
	{
		out << "\nstruct finalization { ~finalization(); };" << newline();
		out << "finalization::~finalization()";
		write_dll_unloader();
		out << newline();
		me.final = true;
	}
	if (!lib)
	{
		out << "\n\nstruct __init_and_final\n{\n";
		table<module*>::iterator i;
		for (i = modules.begin(); i != modules.end(); i++)
		{
			if ((*i)->init)
				out << "\t" << (*i)->name << "::initialization __init_" << (*i)->name << ";\n";
			if ((*i)->final)
				out << "\t" << (*i)->name << "::finalization __final_" << (*i)->name << ";\n";
		}
		out << "} *__init_and_final_ptr;\n\n";
		out << "extern \"C\" void __init_all() { __init_and_final_ptr = new __init_and_final; }\n";
		out << "extern \"C\" void __final_all() { delete __init_and_final_ptr; }\n";
	}
	out << "\n\n}\n}";
}


udt* parser::parse_funccall(function* func, bool lvalue)
{
	if (!func)
	{
		try
		{
			func = (function*)findid(lex.token);
		}
		catch (out_of_range)
		{
			throw "undeclared identifier '" + lex.token + "'";
		}
	}
	lex.next();
	out << "(";
	table<variable*>::iterator i = func->args.begin();
	if (lex.token == "(" || (lvalue && lex.tokkind != lexer::nl))
	{
		if (lex.token == "(")
			lex.next();
		bool first = true;
		while (true)
		{
			if (first)
				first = false;
			else
				lex.next();
			bool byval = false;
			if (lex.token ^ "BYVAL")
			{
				if (!((*i)->type->name ^ "ANY") && !((*i)->type->name ^ "STRING") && !((*i)->type->name ^ "APISTRING"))
					throw "only STRING and ANY arguments can be passed BYVAL";
				byval = true;
				lex.next();
			}
			if (i != func->args.begin())
				out << ", ";
			if (!byval && !(*i)->byref && !((*i)->type->name ^ "ANY"))
			{
				out << "bpp::conv((";
				if ((*i)->array)
					out << "bpp::array<";
				if ((*i)->col == variable::coll)
					out << "bpp::collection<";
				else if ((*i)->col == variable::dict)
					out << "bpp::dictionary<";
				if ((*i)->type->kind == udt::cls)
					out << "bpp::ref<class ";
				out << (*i)->type->cname();
				if ((*i)->col)
					out << "> ";
				if ((*i)->array)
					out << "> ";
				if ((*i)->type->kind == udt::cls)
					out << "> ";
				out << "*)0, ";
			}
			if (lex.token == ",")
			{
				if ((*i)->defval == "")
					throw "argument not optional";
				else
					out << (*i)->defval;
			}
			else
			{
				udt* type;
				if ((*i)->type->name ^ "ANY")
				{
					if (byval)
					{
						out << "(void*)bpp::conv((int*)0, ";
						parse_expr();
						out << ")";
					}
					else
					{
						if (!func->generic)
							out << "&";
						try
						{
							type = parse_expr(true, true);
						}
						catch (invalid_argument)
						{
							throw "not an lvalue";
						}
					}
				}
				else if ((*i)->byref)
				{
					if ((*i)->array || (*i)->col)
					{
						try
						{
							if (parse_expr(true, true, true) != (*i)->type)
								throw "array must be of the same type as argument";
						}
						catch (invalid_argument)
						{
							if ((*i)->array)
								throw "array expected";
							else if ((*i)->col == variable::coll)
								throw "collection expected";
							else if ((*i)->col == variable::dict)
								throw "dictionary expected";
						}
					}
					else
					{
						try
						{
							type = parse_expr(true, true);
						}
						catch (invalid_argument)
						{
							throw "not an lvalue";
						}
						if (type != (*i)->type)
							throw "type mismatch";
					}
				}
				else if (byval && ((*i)->type->name ^ "STRING" || (*i)->type->name ^ "APISTRING"))
				{
					out << "(char*)bpp::conv((int*)0, ";
					parse_expr();
					out << ")";
				}
				else
					type = parse_expr();
				if ((*i)->type->kind == udt::struc || type->kind == udt::struc)
					if ((*i)->type->name != "ANY")
						if (type != (*i)->type)
							throw "type mismatch";
			}
			if (!byval && !(*i)->byref && !((*i)->type->name ^ "ANY"))
				out << ")";
			i++;
			if (lex.token == ")")
				break;
			else if (lex.token != ",")
				if (lvalue)
					break;
				else
					throw "comma or closing brace expected";
			if (i == func->args.end())
				break;
		}
		if (!lvalue)
		{
			if (lex.token != ")")
				throw "closing brace expected";
		}
		if (lex.token == ")")
			lex.next();
	}
	for (bool first = (i == func->args.begin()); i != func->args.end(); i++)
	{
		if ((*i)->defval == "")
			throw "argument not optional";
		if (first)
			first = false;
		else
			out << ", ";
		out << (*i)->defval;
	}
	out << ")";
	return func->type;
}


udt* parser::parse_expr(bool lvalue, bool dma, bool array, bool col)
{
	return parse_expr(lvalue, dma, array, col, &out);
}


udt* parser::parse_expr(bool lvalue, bool dma, bool array, bool col, ostream* os)
{
	ostream& out = *os;
	bool expr = false, first = true;
	string op; ident* id; udt* type;
	while (true)
	{
		if (first)
			first = false;
		else
			out << " ";
		if (lex.token == "-" || lex.token == "+")
		{
			if (lvalue)
				throw invalid_argument("not an lvalue");
			out << lex.token;
			expr = true;
			lex.next();
		}
		else if (lex.token ^ "NOT")
		{
			if (lvalue)
				throw invalid_argument("not an lvalue");
			out << "!";
			expr = true;
			lex.next();
		}
		else if (lex.token ^ "[")
		{
			if (lvalue)
				throw invalid_argument("not an lvalue");
			lex.next();
			if (lex.token != "NOT")
				throw "NOT expected";
			out << "~";
			expr = true;
			lex.next();
			if (lex.token != "]")
				throw "closing brace expected";
			lex.next();
		}
		if (lex.token == "(")
		{
			if (lvalue)
				throw invalid_argument("not an lvalue");
			out << "( ";
			lex.next();
			type = parse_expr(lvalue);
			if (lex.token != ")")
				throw "brackets mismatch";
			out << " )";
			lex.next();			
		}
		else if (lex.tokkind == lexer::num)
		{
			if (lvalue)
				throw invalid_argument("not an lvalue");
			out << lex.token;
			type = findtype("DOUBLE");
			lex.next();
		}
		else if (lex.tokkind == lexer::str)
		{
			if (lvalue)
				throw invalid_argument("not an lvalue");
			out << "string(" << lex.token << ")";
			type = findtype("STRING");
			lex.next();
		}
		else if (lex.token ^ "NEW")
		{
			if (lvalue)
				throw invalid_argument("not an lvalue");
			lex.next();
			type = parse_typename();
			if (type->kind != udt::cls)
				throw "class expected";
			out << "bpp::ref<class " << type->cname()
				<< ">(new (class " << type->cname() << "))";
		}
		else if (lex.token ^ "LBOUND" || lex.token ^ "UBOUND")
		{
			if (lvalue)
				throw invalid_argument("not an lvalue");
			if (lex.token ^ "LBOUND")
				out << "bpp::lbound(";
			else
				out << "bpp::ubound(";
			lex.next();
			if (lex.token != "(")
				throw "brace expected";
			lex.next();
			parse_expr(false, false, true);
			if (lex.token == ",")
			{
				out << ", ";
				lex.next();
				parse_expr();
			}
			if (lex.token != ")")
				throw "closing brace expected";
			out << ")";
			lex.next();
			type = findtype("LONG");
		}
		else if (lex.token ^ "SIZEOF")
		{
			if (lvalue)
				throw invalid_argument("not an lvalue");
			out << "sizeof(";
			lex.next();
			if (lex.token == "(")
			{
				lex.next();
				parse_expr();
				if (lex.token != ")")
					throw "closing brace expected";
				lex.next();
			}
			else
			{
				udt* type = parse_typename();
				if (type->kind == ident::cls)
					out << "bpp::ref<class ";
				out << type->cname();
				if (type->kind == ident::cls)
					out << ">";
			}
			out << ")";
			type = findtype("DWORD");
		}
		else if (lex.token ^ "ADDRESSOF")
		{
			lex.next();
			try
			{
				function* func = (function*)findid(lex.token, &me);
				if (func->kind != ident::func)
					throw out_of_range("");
				out << "((unsigned long)" << func->cname() << ")";
			}
			catch (out_of_range)
			{
				throw "global function expected";
			}
			lex.next();
			type = findtype("DWORD");
		}
		else if (lex.tokkind == lexer::id)
		{
			bool inherited = false;
			string token;
			if (lex.token ^ "SUPER")
			{
				if (!inside)
					throw "SUPER is only legal in methods";
				lex.next();
				if (lex.token != ".")
					throw "dot expected";
				lex.next();
				token = lex.token;
				clas* cls = inside->super;
				while (cls)
				{
					try
					{
						id = cls->methods[token];
						break;
					}
					catch (out_of_range)
					{
						try
						{
							id = cls->props[token];
							variable* prop = (variable*)id;
							if (!prop->propset && !prop->propget)
								throw "property get or set expected";
							break;
						}
						catch (out_of_range)
						{
							cls = cls->super;
						}
					}
				}
				if (!cls)
					throw "undeclared identifier '" + token + "'";
				if (id->scope == ident::priv)
					throw "private member '" + id->name +
						"' is closed for class descendants";
				out << cls->cname() << "::";
				type = cls;
				inherited = true;
			}
			else
			{
				try 
				{
					bool found = false;
					try
					{
						me.idents[lex.token];
						found = true;
					}
					catch (out_of_range)
					{
					}
					if (found)
						throw out_of_range("");
					module* mod = modules[lex.token];
					lex.next();
					if (lex.token != ".")
						throw "dot expected";
					lex.next();
					try
					{
						id = mod->idents[lex.token];
					}
					catch (out_of_range)
					{
						throw "undeclared identifier '" + lex.token + "'";
					}
				}
				catch (out_of_range)
				{
					try
					{
						id = findid(lex.token);
					}
					catch (out_of_range)
					{
						if (lvalue)
						{
							if (explic)
								throw "undeclared identifier '" +
									lex.token + "'";
							else
							{
								id = new variable(lex.token, ident::local, level,
									findtype("VARIANT"));
								me.idents.push_back(id);
								out << "variant ";
							}
						}
						else
							throw "undeclared identifier '" + lex.token + "'";
					}
				}
				if (id->kind == ident::cons)
				{
					constant* cons = (constant*)id;
					if (cons->value[0] == '"')
						out << "string(";
					out << cons->value;
					if (cons->value[0] == '"')
						out << ")";
					type = findtype("VARIANT");
					lex.next();
				}
				else if (id->kind == ident::var)
				{
					out << id->cname();
					type = ((variable*)id)->type;
					lex.next();
					variable* var = (variable*)id;
					if (var->propget || var->propset)
						var->statar = 0;
					if (var->array || var->statar)
					{
						if (lex.token != "(")
							throw "brace expected";
						if (var->array)
							out << "(";
						lex.next();
						int i = 0, max = (var->array ? 4 : var->statar);
						while (lex.token != ")")
						{
							if (++i > max)
								throw "too many indices";
							if (var->statar)
								out << "[";
							parse_expr();
							if (var->statar)
								out << "]";
							if (lex.token ^ ",")
							{
								if (var->array)
									out << ", ";
								lex.next();
							}
							else if (lex.token != ")")
								throw "brackets mismatch";
						}
						if (var->statar)
							if (i != var->statar)
								throw "dimensions mismatch";
						if (var->array)
							out << ")";
						if (array && i)
							expr = true;
						else if (!array && !i)
							throw "specify at least one index";
						lex.next();
					}
					else if (var->col)
					{
						if (lex.token == "(")
						{
							out << "(";
							lex.next();
							if (parse_expr()->kind == ident::struc)
								throw "index should be a scalar";
							if (lex.token != ")")
								throw "closing brace expected";
							out << ")";
							lex.next();
							expr = true;
						}
						else if (lex.token == ".")
						{
							out << ".";
							lex.next();
							if (lex.tokkind != lexer::id)
								throw "identifier expected";
							type = NULL;
							if ((lex.token ^ "Add") || (lex.token ^ "Insert"))
							{
								if (var->col != variable::coll)
									throw "ADD cannot be used on dictionaries";
								out << "add(bpp::conv((";
								if (var->type->kind == udt::cls)
									out << "bpp::ref<class ";
								out << var->type->cname();
								if (var->type->kind == udt::cls)
									out << "> ";
								out << "*)0, ";
								lex.next();
								if (lex.token != "(")
									throw "brace expected";
								lex.next();
								udt* itype = parse_expr();
								if (var->type->kind == ident::struc && var->type != itype)
									throw "type mismatch";
								if (lex.token == ",")
								{
									out << "), bpp::conv((int*)0, ";
									lex.next();
									if (parse_expr()->kind == ident::struc)
										throw "index should be a scalar";
								}
								if (lex.token != ")")
									throw "closing brace expected";
								out << "))";
								lex.next();
							}
							else if (lex.token ^ "Contains")
							{
								if (var->col != variable::dict)
									throw "CONTAINS cannot be used on collections";
								out << "contains(bpp::conv((bpp::string*)0, ";
								lex.next();
								if (lex.token != "(")
									throw "brace expected";
								lex.next();
								if (parse_expr()->kind == ident::struc)
									throw "index should be a scalar";
								if (lex.token != ")")
									throw "closing brace expected";
								out << "))";
								lex.next();
								type = findtype("BOOLEAN");
							}
							else if (lex.token ^ "Remove")
							{
								out << "remove(bpp::conv(";
								if (var->col == variable::coll)
									out << "(int*)0, ";
								else
									out << "(bpp::string*)0, ";
								lex.next();
								if (lex.token != "(")
									throw "brace expected";
								lex.next();
								if (parse_expr()->kind == ident::struc)
									throw "index should be a scalar";
								if (lex.token != ")")
									throw "closing brace expected";
								out << "))";
								lex.next();
							}
							else if (lex.token ^ "RemoveItem")
							{
								out << "removeitem(bpp::conv((";
								if (var->type->kind == udt::cls)
									out << "bpp::ref<class ";
								out << var->type->cname();
								if (var->type->kind == udt::cls)
									out << "> ";
								out << "*)0, ";
								lex.next();
								if (lex.token != "(")
									throw "brace expected";
								lex.next();
								if (parse_expr()->kind == ident::struc)
									throw "index should be a scalar";
								if (lex.token != ")")
									throw "closing brace expected";
								lex.next();
								out << "))";
							}
							else if (lex.token ^ "Count")
							{
								out << "length()";
								type = findtype("DWORD");
								lex.next();
							}
							else if (lex.token ^ "Clear")
							{
								out << "clear()";
								type = NULL;
								lex.next();
							}
							else
								throw "undefined method '" + lex.token + "'";
							if (lvalue)
								throw invalid_argument("not an lvalue");
							else if (!type)
								throw "SUB calls cannot be used in expressions";
							expr = true;
						}
						else if (!col)
							throw "index expected";
					}
				}
				else if (id->kind == ident::func)
				{
					out << id->cname();
					type = parse_funccall((function*)id, lvalue);
					if (!type)
						if (lvalue)
							throw invalid_argument("not an lvalue");
						else
							throw "SUB calls cannot be used in expressions";
				}
				else
					throw "variable or function call expected";
			}
			if (type && type->kind == ident::cls && lex.token ^ "IS")
			{
				lex.next();
				bool child = false;
				if (lex.token == "[" || lex.token == "(")
					child = true, lex.next();
				clas* cls = (clas*) parse_typename();
				if (child)
					if (lex.token != "]" && lex.token != ")")
						throw "square brace expected";
					else
						lex.next();
				if (cls->kind != ident::cls)
					throw "class expected";
				if (child)
					out << "->__isbase";
				else
					out << "->__isa";
				out << "(typeid(" << cls->cname() << "))";
				type = findtype("BOOLEAN");
			}
			while (type && lex.token == "." || inherited)
			{
				if (!inherited)
				{
					lex.next();
					if (id->kind == ident::var)
					{
						variable* var = (variable*)id;
						if (var->propset && !var->propget)
							throw "property '" + var->name +
								"' is write-only";
						if (var->propget)
							out << "()";
					}
					if (type->kind != udt::struc && type->kind != udt::cls)
						throw "'" + type->name + "' is not an aggregate type";
					if (lex.tokkind != lexer::id)
						throw "identifier expected";
					try
					{
						if (type->kind == udt::struc)
						{
							id = ((structure*)type)->fields[lex.token];
							out << ".";
						}
						else
						{
							id = ((clas*)type)->member(lex.token);
							out << "->";
						}
					}
					catch (out_of_range)
					{
						throw "undeclared identifier '" + lex.token + "'";
					}
					if (type->kind == udt::cls)
					{
						if (inside->childof((clas*)type))
						{
							if (id->scope == ident::priv)
							{
								try
								{
									inside->member(id->name, false);
								}
								catch (out_of_range)
								{
									throw "private member '" + id->name +
										"' is closed for class descendants";
								}
							}
						}
						else
							if (id->scope != ident::pub)
								throw "class member  '" + id->name +
									"' is not public";
					}
				}
				else
					inherited = false;
				out << id->cname(false);
				if (id->kind == ident::func)
					parse_funccall((function*)id);
				else if (id->kind == ident::var)
				{
					variable* var = (variable*)id;
					type = var->type;
					if (var->propget || var->propset)
						var->statar = 0;
					lex.next();
					if (var->array || var->statar)
					{
						if (lex.token != "(")
							throw "brace expected";
						if (var->array)
							out << "(";
						lex.next();
						int i = 0, max = (var->array ? 4 : var->statar);
						while (lex.token != ")")
						{
							if (++i > max)
								throw "too many indices";
							if (var->statar)
								out << "[";
							parse_expr();
							if (var->statar)
								out << "]";
							if (lex.token ^ ",")
							{
								if (var->array)
									out << ", ";
								lex.next();
							}
							else if (lex.token != ")")
								throw "brackets mismatch";
						}
						if (var->statar)
							if (i != var->statar)
								throw "dimensions mismatch";
						if (var->array)
							out << ")";
						if (array && i)
							expr = true;
						else if (!array && !i)
							throw "specify at least one index";
						lex.next();
					}
					else if (var->col)
					{
						if (lex.token == "(")
						{
							out << "(";
							lex.next();
							if (parse_expr()->kind == ident::struc)
								throw "index should be a scalar";
							if (lex.token != ")")
								throw "closing brace expected";
							out << ")";
							lex.next();
							expr = true;
						}
						else if (lex.token == ".")
						{
							out << ".";
							lex.next();
							if (lex.tokkind != lexer::id)
								throw "identifier expected";
							type = NULL;
							if ((lex.token ^ "Add") || (lex.token ^ "Insert"))
							{
								if (var->col != variable::coll)
									throw "ADD cannot be used on dictionaries";
								out << "add(bpp::conv((";
								if (var->type->kind == udt::cls)
									out << "bpp::ref<class ";
								out << var->type->cname();
								if (var->type->kind == udt::cls)
									out << "> ";
								out << "*)0, ";
								lex.next();
								if (lex.token != "(")
									throw "brace expected";
								lex.next();
								udt* itype = parse_expr();
								if (var->type->kind == ident::struc && var->type != itype)
									throw "type mismatch";
								if (lex.token == ",")
								{
									out << "), bpp::conv((int*)0, ";
									lex.next();
									if (parse_expr()->kind == ident::struc)
										throw "index should be a scalar";
								}
								if (lex.token != ")")
									throw "closing brace expected";
								lex.next();
								out << "))";
							}
							else if (lex.token ^ "Contains")
							{
								if (var->col != variable::dict)
									throw "CONTAINS cannot be used on collections";
								out << "contains(bpp::conv((bpp::string*)0, ";
								lex.next();
								if (lex.token != "(")
									throw "brace expected";
								lex.next();
								if (parse_expr()->kind == ident::struc)
									throw "index should be a scalar";
								if (lex.token != ")")
									throw "closing brace expected";
								out << "))";
								lex.next();
								type = findtype("BOOLEAN");
							}
							else if (lex.token ^ "Remove")
							{
								out << "remove(bpp::conv(";
								if (var->col == variable::coll)
									out << "(int*)0, ";
								else
									out << "(bpp::string*)0, ";
								lex.next();
								if (lex.token != "(")
									throw "brace expected";
								lex.next();
								udt* itype = parse_expr();
								if (var->type->kind == ident::struc && var->type != itype)
									throw "type mismatch";
								if (lex.token != ")")
									throw "closing brace expected";
								lex.next();
								out << "))";
							}
							else if (lex.token ^ "RemoveItem")
							{
								out << "removeitem(bpp::conv((";
								if (var->type->kind == udt::cls)
									out << "bpp::ref<class ";
								out << var->type->cname();
								if (var->type->kind == udt::cls)
									out << "> ";
								out << "*)0, ";
								lex.next();
								if (lex.token != "(")
									throw "brace expected";
								lex.next();
								if (parse_expr()->kind == ident::struc)
									throw "index should be a scalar";
								if (lex.token != ")")
									throw "closing brace expected";
								lex.next();
								out << "))";
							}
							else if (lex.token ^ "Count")
							{
								out << "length()";
								type = findtype("INTEGER");
								lex.next();
							}
							else if (lex.token ^ "Clear")
							{
								out << "clear()";
								type = NULL;
								lex.next();
							}
							else
								throw "undefined method '" + lex.token + "'";
							if (lvalue)
								throw invalid_argument("not an lvalue");
							else if (!type)
								throw "SUB calls cannot be used in expressions";
							expr = true;
						}
						else if (!col)
							throw "index expected";
					}
//					lex.next();
				}
			}
			if (id->kind == ident::var)
			{
				variable* var = (variable*)id;
				if (var->propset && !var->propget && !lvalue)
					throw "property '" + var->name +
						"is write-only";
				if (var->propget && !lvalue)
					out << "()";
			}
			if (lvalue && id->kind == ident::func)
				throw invalid_argument("not an lvalue");
			if (array)
			{
				if (id->kind != ident::var)
					throw "array expected";
				if (!((variable*)id)->array)
					throw "array expected";
			}
			if (col)
			{
				if (id->kind != ident::var)
					throw "collection expected";
				if (!((variable*)id)->col)
					throw "collection expected";
			}
		}
		else
			throw "expression expected";
		if (type->kind == udt::struc)
		{
			if (expr)
				throw "structures cannot be used in expressions";
			break;
		}
		if (array)
		{
			if (expr)
				throw "array expected";
			break;
		}
		if (col)
		{
			if (expr)
				throw "collection expected";
			break;
		}
		if (lex.token == "+" || lex.token == "-" || lex.token == "*" ||
			lex.token == "/" || lex.token == ">" || lex.token == "<" ||
			lex.token == ">=" || lex.token == "<=")
		{
			out << " " << lex.token;
		}
		else if (lex.token == "<>")
			out << " !=";
		else if (lex.token ^ "MOD")
			out << " %";
		else if (lex.token ^ "AND")
			out << " &&";
		else if (lex.token ^ "OR")
			out << " ||";
		else if (lex.token ^ "XOR")
			out << " ^";
		else if (lex.token == "[")
		{
			lex.next();
			if (lex.token ^ "AND")
				out << " &";
			else if (lex.token ^ "OR")
				out << " |";
			else
				throw "AND or OR expected";
			lex.next();
			if (lex.token != "]")
				throw "closing brace expected";
		}
		else if (lex.token == "=" && !lvalue)
			out << " ==";
		else
			break;
		if (lvalue)
			throw invalid_argument("not an lvalue");
		lex.next();
	}
	if (lvalue && !dma)
	{
		variable* var = (variable*)id;
		if (type->name ^ "EVENT")
		{
			out << ".set";
		}
		else
		{
			if (!var->propset)
			{
				if (var->propget)
					throw "property '" + var->name + "' is read-only";
				if (lex.token == "=")
					out << " = ";
				else if (lex.token == "+=")
					out << " += ";
				else if (lex.token == "-=")
					out << " -= ";
				else if (lex.token == "*=")
					out << " *= ";
				else if (lex.token == "/=")
					out << " /= ";
				else if (lex.token == "++")
					out << "++";
				else if (lex.token == "--")
					out << "--";
			}
			else if (lex.token != "=")
				throw "this operation is not supported on properties";
		}
	}
	if (lvalue && !dma && lex.token != "++" && lex.token != "--")
	{
		out << "(";
		out << "bpp::conv((";
		if (type->kind == udt::cls)
			out << "bpp::ref<class ";
		out << type->cname();
		if (type->kind == udt::cls)
			out << "> ";
		out << "*)0, ";
	}
	return type;
}


udt* parser::parse_typename()
{
	udt *result;
	string token = lex.token;
	if (lex.tokkind != lexer::id)
		throw "identifier expected";
	lex.next();
	try 
	{
		if (lex.token ^ ".")
		{
			lex.next();
			if (lex.tokkind != lexer::id)
				throw "identifier expected";
			module* mod;
			try
			{
				mod = modules[token];
			}
			catch (out_of_range)
			{
				throw "module '" + token + "'not found";
			}
			result = findtype(lex.token, mod);
			lex.next();
		}
		else
			result = findtype(token);
	}
	catch (out_of_range)
	{
		throw "undefined type";
	}
	return result;
}


void parser::parse_dim(int scope)
{
	if (lex.token ^ "DIM")
		lex.next();
	while (true)
	{
		variable* var = new variable("", scope, level);
		vector<string> vars;
		if (lex.token == "(")
		{
			lex.next();
			while (true)
			{
				if (lex.tokkind != lexer::id)
					throw "identifier expected";
				try
				{
					ident* id = me.idents[lex.token];
					if (id->kind == ident::var)
						if (((variable*)id)->level < level)
							throw out_of_range("");
						throw "duplicate identifier '" + me.idents[lex.token]->name + "'";
				}
				catch (out_of_range)
				{
				}
				validate(lex.token);
				vars.push_back(lex.token);
				lex.next();
				if (lex.token != ",")
					break;
				lex.next();
			}
			if (lex.token != ")")
				throw "closing brace expected";
			lex.next();
		}
		else
		{
			if (lex.tokkind != lexer::id)
				throw "identifier expected";
			try
			{
				ident* id = me.idents[lex.token];
				if (id->kind == ident::var)
					if (((variable*)id)->level < level)
						throw out_of_range("");
				throw "duplicate identifier '" + me.idents[lex.token]->name + "'";
			}
			catch (out_of_range)
			{
				validate(lex.token);
				vars.push_back(lex.token);
				lex.next();
			}
		}
		ostringstream os;
		if (lex.token == "(")
		{
			var->array = true;
			os << "(";
			int i = 0;
			lex.next();
			while (lex.token != ")")
			{
				if (++i > 4)
					throw "too many indices";
				parse_expr(false, false, false, false, &os);
				if (lex.token ^ "TO")
				{
					lex.next();
					os << ", ";
					parse_expr(false, false, false, false, &os);
				}
				else
					os << ", 0";
				if (lex.token ^ ",")
				{
					os << ", ";
					lex.next();
				}
				else if (lex.token != ")")
					throw "brackets mismatch";
			}
			os << ")";
			lex.next();
		}
		bool autonew = false;
		if (lex.token ^ "AS")
		{
			lex.next();
			if (lex.token ^ "NEW")
			{
				autonew = true;
				lex.next();
			}
			var->type = parse_typename();
			if (autonew && var->type->kind != ident::cls)
				throw "class name expected";
		}
		else
			var->type = findtype(var->deftype());
		if (var->type->scope < scope)
			throw "cannot use private type in public declaration";
		if (lex.token ^ "COLLECTION")
		{
			if (var->array)
				throw "array of collections is not allowed";
			var->col = variable::coll;
			lex.next();
		}
		else if (lex.token ^ "DICTIONARY")
		{
			if (var->array)
				throw "array of dictionaries is not allowed";
			var->col = variable::dict;
			lex.next();
		}
		if (var->array)
			out << "bpp::array<";
		if (var->col == variable::coll)
			out << "bpp::collection<";
		else if (var->col == variable::dict)
			out << "bpp::dictionary<";
		if (var->type->kind == udt::cls)
			out << "bpp::ref<class ";
		out << var->type->cname();
		if (var->type->kind == udt::cls)
			out << "> ";
		if (var->array)
			out << "> ";
		if (var->col)
			out << "> ";
		out << " ";
		vector<string>::iterator i;
		for (i = vars.begin(); i != vars.end(); i++)
		{
			variable* v = new variable(*i, var->scope, var->level, var->type);
			v->array = var->array;
			v->col = var->col;
			if (i != vars.begin())
				out << ", ";
			out << v->cname();
			if (os.str() != "()")
				out << os.str();
			me.idents.push_back(v);
		}
		if (autonew)
		{
			if (vars.size() != 1)
				throw "cannot initialize groups";
			out << " = bpp::ref<class " << var->type->cname()
				<< ">(new (class " << var->type->cname() << "))";
		}
		else if (lex.token == "=")
		{
			if (vars.size() != 1)
				throw "cannot initialize groups";
			if (var->type->kind == udt::struc)
				throw "structures cannot be initialized";
			if (var->array)
				throw "arrays cannot be initialized";
			out << " = bpp::conv((";
			if (var->type->kind == udt::cls)
				out << "bpp::ref<class ";
			out << var->type->cname();
			if (var->type->kind == udt::cls)
				out << "> ";
			out << "*)0, ";
			lex.next();
			parse_expr();
			out << ")";
		}
		out << "; ";
		if (lex.token ^ ",")
		{
			lex.next();
			continue;
		}
		else
			return;
	}
}


void parser::parse_type(int scope)
{
	lex.next();
	if (lex.tokkind != lexer::id)
		throw "identifier expected";
	try
	{
		throw "duplicate identifier '" +
			me.types[lex.token]->cname() + "'";
	}
	catch (out_of_range)
	{
	}
	structure* struc = new structure(lex.token, scope);
	validate(struc->name);
	me.types.push_back(struc);
	out << "struct " << struc->name << newline() << "{";
	indent++;
	out << newline();
	lex.next();
	if (lex.tokkind != lexer::nl)
		throw "end of statement expected";
	while (true)
	{
		lex.next();
		if (lex.tokkind == lexer::nl)
			continue;
		if (lex.tokkind != lexer::id)
			throw "identifier expected";
		if (lex.token ^ "END")
		{
			lex.next();
			if (!(lex.token ^ "TYPE"))
				throw "TYPE expected";
			lex.next();
			indent--;
			out << newline() << "};";
			break;
		}
		try
		{
			throw "duplicate identifier '" +
				struc->fields[lex.token]->cname() + "'";
		}
		catch (out_of_range)
		{
		}
		variable* field = new variable(lex.token, ident::pub);
		lex.next();
		int dims[6] = { 0 };
		if (lex.token == "(")
		{
			int i = 0;
			while (true)
			{
				if (i > 4)
					throw "too many dimensions";
				lex.next();
				bool str;
				dims[i] = int(atof(parse_conexpr(str).c_str())) + 1;
				if (str)
					throw "integer constant expected";
				field->stdims[i] = dims[i];
				i++;
				if (lex.token == ")")
					break;
				else if (lex.token != ",")
					throw "comma or closing brace expected";
			}
			dims[i] = field->stdims[i] = 0;
			field->statar = i;
			lex.next();
		}
		if (lex.token ^ "AS")
		{
			lex.next();
			field->type = parse_typename();
		}
		else
			field->type = findtype(field->deftype());
		if (field->type->scope < scope)
			throw "cannot use private type in public declaration";
		int strl = 0;
		if (lex.token == "*" && (field->type->name ^ "STRING" || field->type->name ^ "APISTRING"))
		{
			lex.next();
			bool str;
			strl = int(atof(parse_conexpr(str).c_str()));
			if (str)
				throw "integer constant expected";
			if (strl < 1)
				throw "invalid string length";
		}
		struc->fields.push_back(field);
		if (field->type->kind == udt::cls)
			out << "bpp::ref<class ";
		if (strl)
			out << "bpp::flstring<" << strl << "> ";
		else
			out << field->type->cname();
		if (field->type->kind == udt::cls)
			out << ">";
		out << " " << field->cname();
		for (int i = 0; i < 5; i++)
			if (dims[i])
				out << "[" << dims[i] << "]";
			else
				break;
		out << ";" << newline();
		if (lex.tokkind != lexer::nl && lex.token != ",")
			throw "comma or end of statement expected";
	}
}


string parser::parse_conexpr(bool& str)
{
	enum { set, add, sub, mul, div, and, or, xor } oper = set;
	string result;
	double nres;
	while (true)
	{
		bool str2, neg = false, not = false;
		string op;
		double nop;
		if (lex.token == "+")
			lex.next();
		else if (lex.token == "-")
		{
			neg = true;
			lex.next();
		}
		else if (lex.token ^ "NOT")
		{
			not = true;
			lex.next();
		}
		switch (lex.tokkind)
		{
		case lexer::num:
			str2 = false;
			op = lex.token;
			lex.next();
			break;
		case lexer::str:
			str2 = true;
			op = lex.token.substr(1, lex.token.size() - 2);
			lex.next();
			break;
		case lexer::id:
		{
			constant* cons;
			string token = lex.token;
			lex.next();
			try
			{
				if (lex.token == ".")
				{
					lex.next();
					if (lex.tokkind != lexer::id)
						throw "identifier expected";
					cons = (constant*)modules[token]->idents[lex.token];
					lex.next();
				}
				else
				{
					cons = (constant*)findid(token);
				}
			}
			catch (out_of_range)
			{
				throw "undeclared identifier '" + token + "'";
			}
			if (cons->kind != ident::cons)
				throw "constant expected";
			str2 = cons->str;
			op = cons->value;
			if (str2)
				op = op.substr(1, op.size() - 2);
			break;
		}
		case lexer::sym:
			if (lex.token == "(")
			{
				lex.next();
				op = parse_conexpr(str2);
				if (lex.token != ")")
					throw "closing brace expected";
				lex.next();
				if (str2)
					op = op.substr(1, op.size() - 2);
			}
			else
				throw "constant expression expected";
			break;
		default:
			throw "constant expected";
		}
		nop = atof(op.c_str());
		if (neg)
			nop = -nop;
		else if (not)
			nop = ~int(nop);
		switch (oper)
		{
		case set:
			str = str2;
			if (str)
				result = op;
			else
				nres = nop;
			break;
		case add:
			if (str)
				result += op;
			else
				nres += nop;
			break;
		case sub:
			if (str)
				throw "cannot subtract strings";
			else
				nres -= nop;
			break;
		case mul:
			if (str)
				throw "cannot multiply strings";
			else
				nres *= nop;
			break;
		case div:
			if (str)
				throw "cannot divide strings";
			else
				nres /= nop;
			break;
		case and:
			if (!str)
				nres = int(nres) & int(nop);
			break;
		case or:
			if (!str)
				nres = int(nres) | int(nop);
			break;
		case xor:
			if (!str)
				nres = int(nres) ^ int(nop);
			break;
		}
		if (lex.token == "+")
			oper = add;
		else if (lex.token == "-")
			oper = sub;
		else if (lex.token == "*")
			oper = mul;
		else if (lex.token == "/")
			oper = div;
		else if (lex.token ^ "AND")
			oper = and;
		else if (lex.token ^ "OR")
			oper = or;
		else if (lex.token ^ "XOR")
			oper = xor;
		else
			break;
		lex.next();
	}
	if (str)
		return '"' + result + '"';
	else
		if (abs(nres) - int(abs(nres)))
			return ftos(nres);
		else
			return itos(nres);
}


void parser::parse_const(int scope)
{
	lex.next();
	if (lex.tokkind != lexer::id)
		throw "identifier expected";
	constant* cons = new constant(lex.token, scope);
	validate(cons->name);
	lex.next();
	if (lex.token ^ "AS")
	{
		lex.next();
		parse_typename();
	}
	if (lex.token != "=")
		throw "assignment expected";
	lex.next();
	cons->value = parse_conexpr(cons->str);
	try
	{
		me.idents[cons->name];
	}
	catch (out_of_range)
	{
		me.idents.push_back(cons);
	}
	out << "// const " << cons->name << " = " << cons->value;
}


void parser::parse_block()
{
	while (lex.tokkind != lexer::eof)
	{
		out << newline();
		if (!parse_statement())
			break;
		if (lex.tokkind != lexer::nl)
			throw "end of statement expected";
		lex.next();
	}
}


void parser::parse_exit()
{
	lex.next();
	if (lex.token ^ "DO" || lex.token ^ "FOR")
	{
		if (loops)
			out << "break;";
		else
			throw "EXIT DO/FOR outside of the loop";
	}
	else if (lex.token ^ "FUNCTION" || lex.token ^ "SUB")
	{
		// debug build
		if (devmode)
			out << "bpp::dbg_endfunc();" << newline();
		if (infunc)
			out << "return result;";
		else
			out << "return;";
	}
	else if (lex.token ^ "PROPERTY")
	{
		lex.next();
		if (!(lex.token ^ "GET") && !(lex.token ^ "SET"))
			throw "GET or SET expected";
		// debug build
		if (devmode)
			out << "bpp::dbg_endfunc();" << newline();
		if (infunc)
			out << "return result;";
		else
			out << "return;";
	}
	else
		throw "DO, FOR, FUNCTION, SUB or PROPERTY expected";
	lex.next();
}


void parser::parse_swap()
{
	out << "bpp::swap(";
	lex.next();
	udt* type = parse_expr(true, true);
	out << ", ";
	if (lex.token != ",")
		throw "comma expected";
	lex.next();
	if (parse_expr(true, true) != type)
		throw "type mismatch";
	out << ");";
}


bool parser::parse_statement()
{
	while (lex.tokkind == lexer::nl)
		lex.next();
	if (lex.tokkind == lexer::lab)
	{
		parse_label();
		return true;
	}
	if (lex.tokkind != lexer::id)
		throw "statement expected";
	// debug build
	if (devmode)
		out << "bpp::dbg_line(" << lex.line << "); ";
	if (lex.token ^ "END")
	{
		lex.next();
		if (lex.tokkind != lexer::nl)
			return false;
		out << "throw bpp::end();";
	}
	else if (lex.token ^ "NEXT" || lex.token ^ "WEND" ||
		lex.token ^ "LOOP" || lex.token ^ "ELSE" ||
		lex.token ^ "ELSEIF" || lex.token ^ "CATCH" ||
		lex.token ^ "CASE")
	{
		return false;
	}
	else if (lex.token ^ "EXIT")
		parse_exit();
	else if (lex.token ^ "DIM")
		parse_dim(ident::local);
	else if (lex.token ^ "LET" || lex.token ^ "SET" || lex.token ^ "CALL")
	{
		lex.next();
		parse_setorcall();
	}
	else if (lex.token ^ "IF")
		parse_if();
	else if (lex.token ^ "WHILE")
		parse_while();
	else if (lex.token ^ "DO")
		parse_do();
	else if (lex.token ^ "FOR")
		parse_for();
	else if (lex.token ^ "REDIM")
		parse_redim();
	else if (lex.token ^ "PRINT")
		parse_print();
	else if (lex.token ^ "INPUT")
		parse_input();
	else if (lex.token ^ "LINE")
		parse_lineinput();
	else if (lex.token ^ "RAISE")
		parse_raise();
	else if (lex.token ^ "TRY")
		parse_try();
	else if (lex.token ^ "THROW")
		parse_throw();
	else if (lex.token ^ "SELECT")
		parse_select();
	else if (lex.token ^ "INCR")
		parse_incr();
	else if (lex.token ^ "DECR")
		parse_decr();
	else if (lex.token ^ "SWAP")
		parse_swap();
	else if (lex.token ^ "GOTO")
		parse_goto();
	else
		parse_setorcall();
	return true;
}


void parser::parse_property(int scope, clas* cls)
{
	lex.next();
	if (lex.tokkind != lexer::id)
		throw "identifier expected";
	variable* prop;
	try
	{
		prop = (variable*) cls->member(lex.token);
		if (prop->kind != ident::var)
			throw "duplicate identifier " + prop->name;
		try
		{
			cls->props[lex.token];
			throw "duplicate identifier " + prop->name;
		}
		catch (out_of_range)
		{
			cls->props.push_back(prop);
		}
	}
	catch (out_of_range)
	{
		prop = new variable(lex.token, scope);
		lex.next();
		if (lex.token ^ "AS")
		{
			lex.next();
			prop->type = parse_typename();
		}
		else
			prop->type = findtype("VARIANT");
		prop->propget = prop->propset = true;
		if (lex.token ^ "READONLY")
		{
			prop->propset = false;
			lex.next();
		}
		else if (lex.token ^ "WRITEONLY")
		{
			prop->propget = false;
			lex.next();
		}
		cls->props.push_back(prop);
	}
	if (prop->propget)
	{
		out << "\tvirtual ";
		if (prop->type->kind == udt::cls)
			out << "bpp::ref<class ";
		out << prop->type->cname();
		if (prop->type->kind == udt::cls)
			out << "> ";
		out << " " << prop->cname() << "();" << newline();
	}
	if (prop->propset)
	{
		out << "\tvirtual void " << prop->cname() << "(";
		if (prop->type->kind == udt::cls)
			out << "bpp::ref<class ";
		out << prop->type->cname();
		if (prop->type->kind == udt::cls)
			out << "> ";
		out << ");" << newline();
	}
}


void parser::parse_propgs(int)
{
	bool propset;
	lex.next();
	if (lex.token ^ "GET")
		propset = false;
	else if (lex.token ^ "SET")
		propset = true;
	else
		throw "GET or SET expected";
	lex.next();
	clas* cls = (clas*) findtype(lex.token, &me);
	if (cls->kind != ident::cls)
		throw "class name expected";
	inside = cls;
	lex.next();
	if (lex.token != ".")
		throw "dot expected";
	lex.next();
	try
	{
		variable* prop = cls->props[lex.token];
		if (propset && !prop->propset)
			throw "property is read-only";
		if (!propset && !prop->propget)
			throw "property is write-only";
		lex.next();
		if (lex.tokkind != lexer::nl)
			throw "newline expected";
		enter();
		if (propset)
			me.idents.push_back(new variable(prop->name,
				ident::local, level, prop->type));
		if (propset)
			out << "void";
		else
		{
			if (prop->type->kind == udt::cls)
				out << "bpp::ref<class ";
			out << prop->type->cname();
			if (prop->type->kind == udt::cls)
				out << "> ";
		}
		out << " " << cls->cname() << "::" << prop->cname() << "(";
		if (propset)
		{
			if (prop->type->kind == udt::cls)
				out << "bpp::ref<class ";
			out << prop->type->cname();
			if (prop->type->kind == udt::cls)
				out << "> ";
			out << " " << prop->cname();
		}
		out << ")" << newline() << "{";
		indent++;
		if (!propset)
		{
			out << newline();
			if (prop->type->kind == udt::cls)
				out << "bpp::ref<class ";
			out << prop->type->cname();
			if (prop->type->kind == udt::cls)
				out << "> ";
			out << " " << prop->cname() << ";";
		}
		clas* k = cls;
		while (k)
		{
			{
				table<variable*>::iterator i;
				for	(i = k->props.begin(); i != k->props.end(); i++)
					if (!((*i)->name ^ prop->name))
						me.idents.push_back(*i);
			}
			{
				table<function*>::iterator i;
				for (i = k->methods.begin(); i != k->methods.end(); i++)
					me.idents.push_back(*i);
			}
			k = k->super;
		}
		enter();
		me.idents.push_back(new variable("me", ident::local, level, inside));
		out << newline() << "bpp::ref<class " + inside->cname() +
				"> me(this, true);";
		if (!propset)
			me.idents.push_back(new variable(prop->name,
				ident::local, level, prop->type));
		lex.next();
		parse_block();
		if (!propset)
			out << "return " << prop->cname() << ";";
		indent--;
		leave();
		leave();
		out << newline() << "}";
		if (!(lex.token ^ "PROPERTY"))
			throw "PROPERTY expected";
		lex.next();
		if (propset && !(lex.token ^ "SET"))
			throw "SET expected";
		if (!propset && !(lex.token ^ "GET"))
			throw "GET expected";
		lex.next();
	}
	catch (out_of_range)
	{
		throw "property name expected";
	}
	inside = NULL;
}


void parser::parse_func(int scope, bool proto, clas* cls)
{
	bool noret, external = false, virt = false;
	bool dll = false, export = false;
	string dllname, funcalias;
	if (lex.token ^ "EXPORT")
	{
		if (!this->dll)
			throw "functions can only be exported from DLLs";
		if (scope != ident::pub)
			scope = ident::pub;
		export = true;
		lex.next();
	}
	else if (lex.token ^ "EXTERNAL")
	{
		external = true;
		lex.next();
	}
	if (lex.token ^ "SUB")
		noret = true;
	else if (lex.token ^ "FUNCTION")
		noret = false;
	else
		throw "SUB or FUNCTION expected";
	lex.next();
	if (lex.tokkind != lexer::id)
		throw "identifier expected";
	string token = lex.token;
	function* func;
	lex.next();
	if (lex.token == ".")
	{
		if (cls)
			throw "class specifier is not allowed here";
		try
		{
			cls = (clas*)findtype(token);
		}
		catch (out_of_range)
		{
			throw "undefined class";
		}
		if (cls->kind != udt::cls)
			throw "'" + cls->name + "' is not a class";
		lex.next();
		if (lex.tokkind != lexer::id)
			throw "identifier expected";
		try
		{
			func = cls->methods[lex.token];
		}
		catch (out_of_range)
		{
			throw "'" + lex.token + "' is not a member of '" +
				cls->name + "'";
		}
		inside = cls;
		lex.next();
	}
	else
	{
		try
		{
			if (cls)
			{
				virt = true;
				try
				{
					throw "duplicate identifier '" +
						cls->methods[token]->name + "'";
				}
				catch (out_of_range)
				{
				}
				func = (function*)cls->member(token);
				if (func->kind != ident::func)
					throw "duplicate identifier '" + func->name + "'";
				if (func->scope != scope)
					throw "scope mismatch";
			}
			else
			{
				ident* id = me.idents[token];
				if (id->kind != ident::func)
					throw "duplicate identifier '" + id->name + "'";
				if (proto)
					"function '" + id->name + "' already declared";
				func = (function*)id;
				if (func->resolved)
					throw "function '" + func->name + "'already defined";
			}
			if (noret != (func->type == NULL))
				throw "definition of '" + func->name + "' doesn't "
					"match declaration";
			if (lex.tokkind != lexer::nl)
				throw "function prototype is already defined";
			if (cls)
				cls->methods.push_back(func);
		}
		catch (out_of_range)
		{
			if (cls && token ^ cls->name)
				throw "method cannot have the same name as its class";
			func = new function(token, scope, proto, cls != NULL);
			validate(func->name);
			func->external = external;
			if (lex.token ^ "LIB")
			{
				if (cls)
					throw "method cannot be loaded from DLL";
				proto = dll = true;
				lex.next();
				if (lex.tokkind != lexer::str)
					throw "library name expected";
				dllname = "bpp::string(" + lex.token + ")";
				{
					vector<string>::iterator i;
					bool newlib = true;
					for (i = dlls.begin(); i != dlls.end(); i++)
						if (*i ^ dllname)
							newlib = false;
					if (newlib)
						dlls.push_back(dllname);
				}
				lex.next();
				if (lex.token ^ "ALIAS")
				{
					lex.next();
					if (lex.tokkind != lexer::str)
						throw "function alias expected";
					funcalias = lex.token;
					lex.next();
				}
				else
					funcalias = "bpp::string(\"" + func->name +"\")";
			}
			if (lex.token == "(")
			{
				while (true)
				{
					lex.next();
					bool byref = vbdll;
					if (lex.token ^ "BYREF")
					{
						byref = true;
						lex.next();
					}
					else if (lex.token ^ "BYVAL")
					{
						byref = false;
						lex.next();
					}
					if (lex.tokkind != lexer::id)
						throw "identifier expected";
					try
					{
						throw "duplicate identifier '" +
							func->args[lex.token]->name;
					}
					catch (out_of_range)
					{
					}
					variable* arg = new variable(lex.token, ident::local, level + 1);
					validate(arg->name);
					arg->byref = byref;
					if (!noret && arg->name ^ "result")
						throw "illegal argument name";
					lex.next();
					if (lex.token == "(")
					{
						lex.next();
						if (lex.token != ")")
							throw "closing brace expected";
						arg->byref = arg->array = true;
						lex.next();
					}
					if (lex.token ^ "AS")
					{
						lex.next();
						arg->type = parse_typename();
					}
					else
						arg->type = findtype(arg->deftype());
					if (cls)
					{
						if (arg->type->scope < cls->scope)
							throw "cannot use private type in public declaration";
					}
					else
					{
						if (arg->type->scope < func->scope)
							throw "cannot use private type in public declaration";
					}
					/*if (lex.token ^ "COLLECTION")
					{
						arg->col = variable::coll;
						arg->byref = true;
						lex.next();
					}*/
					if (arg->type->name ^ "ANY")
						arg->byref = false;
					else if (arg->type->kind == ident::struc)
						arg->byref = true;
					else if (arg->type->kind == ident::cls)
						arg->byref = false;
					if (lex.token == "=")
					{
						if (arg->byref)
							throw "reference arguments cannot be optional";
						lex.next();
						bool str;
						arg->defval = parse_conexpr(str);
						if (str)
							arg->defval = "string(" + arg->defval + ")";
					}
					else
						arg->defval = "";
					func->args.push_back(arg);
					if (lex.token == ")")
						break;
					else if (lex.token != ",")
					 	throw "comma or closing brace expected";
				}
				lex.next();
			}
			if (noret)
				func->type = NULL;
			else
				if (lex.token ^ "AS")
				{	
					lex.next();
					func->type = parse_typename();
				}
				else if (lex.tokkind == lexer::nl)
					func->type = findtype(func->deftype());
				else
					throw "end of statement expected";
			if (func->type)
				if (cls)
				{
					if (func->type->scope < cls->scope)
						throw "cannot use private type in public declaration";
				}
				else
				{
					if (func->type->scope < func->scope)
						throw "cannot use private type in public declaration";
				}
			if (cls)
				cls->methods.push_back(func);
			else
				me.idents.push_back(func);
		}
		func->resolved = !proto;
	}
	if (func->external && !dll || export)
		out << "extern \"C\" ";
	if (export)
		out << "__declspec(dllexport) ";
	if (func->scope == ident::priv && !func->method)
		out << "static ";
	if (virt)
		out << "virtual ";
	if (func->type && func->type->kind == udt::cls)
		out << "bpp::ref<class ";
	if (noret)
		out << "void";
	else
		out << func->type->cname();
	if (func->type && func->type->kind == udt::cls)
		out << ">";
	out << " ";
	if (!cls && !func->external)
		out << "__stdcall ";
	if (inside)
		out << inside->cname() << "::";
	out << func->cname(false) << "(";
	table<variable*>::iterator i;
	for (i = func->args.begin(); i != func->args.end(); i++)
	{
		out << (i == func->args.begin() ? "" : ", ");
		if ((*i)->array)
			out << "bpp::array<";
		if ((*i)->col == variable::coll)
			out << "bpp::collection<";
		else if ((*i)->col == variable::dict)
			out << "bpp::dictionary<";
		if ((*i)->type->kind == udt::cls)
			out << "bpp::ref<class ";
		out << (*i)->type->cname();
		if ((*i)->col)
			out << "> ";
		if ((*i)->array)
			out << "> ";
		if ((*i)->type->kind == udt::cls)
			out << ">";
		out << ((*i)->byref ? "& " : " ") << (*i)->cname(false);
	}
	out << ")";
	if (proto)
	{
		if (dll)
		{
			out << newline() << "{";
			indent++; out << newline();
			out << "static void* dll = bpp::getlibhandle("
				<< dllname << ");" << newline();
			out << "static union { void* ptr; ";
			if (func->type)
				if (func->type->kind == ident::cls)
					throw "DLL functions cannot return objects";
				else
					out << func->type->cname();
			else
				out << "void";
			out << " (__stdcall *func)(";
			vector<variable*>::iterator i;
			for (i = func->args.begin(); i != func->args.end(); i++)
			{
				if (i != func->args.begin())
					out << ", ";
				if ((*i)->byref || (*i)->type->name ^ "STRING" || (*i)->type->name ^ "APISTRING")
					out << "void*";
				else
					out << (*i)->type->cname();
			}
			out << "); }; ptr = bpp::getprocaddr(dll, " << funcalias << ");";
			out << newline();
			if (func->type)
				out << "return ";
			out << "func(";
			for (i = func->args.begin(); i != func->args.end(); i++)
			{
				if ((*i)->array || (*i)->col ||
					(*i)->type->kind == ident::cls)
				{
					throw "DLL functions cannot take objects";
				}
				else
				{
					if (i != func->args.begin())
						out << ", ";
					if ((*i)->byref || (*i)->type->name ^ "STRING" || (*i)->type->name ^ "APISTRING")
						out << "&";
					out << (*i)->cname();
				}
			}
			out << ");";
			indent--;
			out << newline() << "}";
		}
		else
			out << ";";
	}
	if (lex.token ^ "GENERIC")
	{
		func->generic = true;
		lex.next();
	}
	if (proto)
		return;
	infunc = !noret;
	labels.clear();
	enter();
	if (inside)
	{
		clas* k = inside;
		while (k)
		{
			{
				table<variable*>::iterator i;
				for	(i = k->props.begin(); i != k->props.end(); i++)
					me.idents.push_back(*i);
			}
			{
				table<function*>::iterator i;
				for (i = k->methods.begin(); i != k->methods.end(); i++)
					me.idents.push_back(*i);
			}
			k = k->super;
		}
	}
	enter();
	for (i = func->args.begin(); i != func->args.end(); i++)
		me.idents.push_back(*i);
	out << newline() << "{";
	// debug build
	if (devmode)
	{
		out << newline() << "bpp::dbg_func(\"" << me.name << ".";
		if (inside)
			out << inside->name << ".";
		out << func->name << "\");" << newline();
	}
	indent++;
	if (!noret)
	{
		me.idents.push_back(new variable("result", ident::local, level, func->type));
		out << newline() << func->type->cname() << " result;";
	}
	if (inside)
	{
		me.idents.push_back(new variable("me", ident::local, level, inside));
		out << newline() << "bpp::ref<class " + inside->cname() +
			"> me(this, true);";
	}
	lex.next();
	parse_block();
	// debug build
	if (devmode)
		out << newline() << "bpp::dbg_endfunc();" << newline();
	if (!noret)
	{
		if (!(lex.token ^ "FUNCTION"))
			throw "FUNCTION expected";
		out << "return result;";
	}
	else
	{
		if (!(lex.token ^ "SUB"))
			throw "SUB expected";
		out << "return;";
	}
	indent--;
	out << newline() << "}" << newline();
	lex.next();
	leave();
	leave();
	{
		map<string, bool>::iterator i;
		for (i = labels.begin(); i != labels.end(); i++)
			if (!i->second)
				throw "undefined label '" + i->first + "'";
	}
	inside = NULL;
}


void parser::parse_setorcall()
{
	try
	{
		udt* type = parse_expr(true);
		if (lex.token == "++" || lex.token == "--")
		{
			out << ";";
			lex.next();
			return;
		}
		if (lex.token != "=" && lex.token != "+=" &&
			lex.token != "-=" && lex.token != "*=" &&
			lex.token != "/=")
		{
			throw "assignment operator expected";
		}
		lex.next();
		if (type->name ^ "EVENT")
		{
			if (!inside)
				throw "must be inside of class to set event handler";
			if (lex.tokkind != lexer::id)
				throw "identifier expected";
			try
			{
				function* handler = inside->method(lex.token);
				if (handler->type || handler->args.size() != 1 ||
					handler->args.front()->type->kind != udt::cls)
				{
					throw "'" + handler->name + "' is not a valid event handler";
				}
				out << "this, &" << inside->cname() << "::"
					<< handler->cname();
			}
			catch (out_of_range)
			{
				throw "undeclared identifier '" + lex.token + "'";
			}
			lex.next();
		}
		else
		{
			if (type != parse_expr(false))
				if (type->kind == udt::struc)
					throw "type mismatch";
		}
		out << " ))";
	}
	catch (invalid_argument)
	{
	}
	out << ";";
}


void parser::parse_if()
{
	out << "if ( ";
	lex.next();
	parse_expr();
	out << " )";
	if (!(lex.token ^ "THEN"))
		throw "THEN expected";
	lex.next();
	enter();
	if (lex.tokkind == lexer::nl)
	{
		out << newline() << "{";
		indent++;
		lex.next();
		parse_block();
		indent--;
		out << newline() << "}";
//		if (!(lex.token ^ "IF"))
//			throw "IF expected";
		while (lex.token ^ "ELSEIF")
		{
			leave();
			out << newline() << "else if ( ";
			lex.next();
			parse_expr();
			out << " ) " << newline() << "{";
			if (!(lex.token ^ "THEN"))
				throw "THEN expected";
			lex.next();
			if (lex.tokkind != lexer::nl)
				throw "end of statement expected";
			lex.next();
			enter();
			indent++;
			parse_block();
			indent--;
			out << newline() << "}";
		}
		if (lex.token ^ "ELSE")
		{
			leave();
			out << newline() << "else" << newline() << "{";
			lex.next();
			if (lex.tokkind != lexer::nl)
				throw "end of statement expected";
			lex.next();
			enter();
			indent++;
			parse_block();
			indent--;
			out << newline() << "}";
		}
		if (!(lex.token ^ "IF"))
			throw "IF expected";
		lex.next();
	}
	else
	{
		out << " { ";
		if (!parse_statement())
			throw "illegal statement";
		out << " }";
	}
	leave();
}


void parser::parse_while()
{
	out << "while ( ";
	lex.next();
	parse_expr();
	out << " )";
	enter();
	if (lex.token ^ "DO")
	{
		out << " { ";
		lex.next();
		if (!parse_statement())
			throw "illegal statement";
		out << " }";
	}
	else if (lex.tokkind == lexer::nl)
	{
		out << newline() << "{";
		indent++; loops++;
		lex.next();
		parse_block();
		indent--; loops--;
		out << newline() << "}";
		if (!(lex.token ^ "WEND"))
			throw "WEND expected";
		lex.next();
	}
	else
		throw "DO or end of statement expected";
	leave();
}


void parser::parse_do()
{
	lex.next();
	if (lex.tokkind == lexer::nl)
	{
		out << "do" << newline() << "{";
		indent++; loops++;
		enter();
		lex.next();
		parse_block();
		leave();
		indent--; loops--;
		if (!(lex.token ^ "LOOP"))
			throw "LOOP expected";
		lex.next();
		out << newline() << "} while ( ";
		if (lex.tokkind == lexer::nl)
			out << "true );";
		else
		{
			if (lex.token ^ "UNTIL")
				out << "!";
			else if (!(lex.token ^ "WHILE"))
				throw "WHILE or UNTIL or end of statement expected";
			out << "( ";
			lex.next();
			parse_expr();
			out << " ) );";
		}
	}
	else
	{
		out << "while ( ";
		if (lex.token ^ "UNTIL")
			out << "!";
		else if (!(lex.token ^ "WHILE"))
			throw "WHILE or UNTIL or end of statement expected";
		out << "( ";
		lex.next();
		parse_expr();
		out << " ) )" << newline() << "{";
		indent++; loops++;
		enter();
		lex.next();
		parse_block();
		leave();
		indent--; loops--;
		if (!(lex.token ^ "LOOP"))
			throw "LOOP expected";
		lex.next();
		out << newline() << "}";
	}
}


void parser::parse_for()
{
	lex.next();
	if (lex.token ^ "EACH")
	{
		parse_foreach();
		return;
	}
	if (lex.tokkind != lexer::id)
		throw "variable expected";
	variable* counter;
	string step = "1";
	bool negative = false;
	out << "// for" << newline() << "{" ;
	enter();
	indent++; loops++;
	out << newline();
	string cntvar = lex.token;
	lex.next();
	if (lex.token ^ "AS")
	{
		lex.next();
		udt* type = parse_typename();
		if (type->kind == udt::struc)
			throw "illegal counter type";
		if (type->kind == udt::cls)
			out << "bpp::ref<class ";
		out << type->cname();
		if (type->kind == udt::cls)
			out << "> ";
		out << " ";
		me.idents.push_back(counter = new variable(cntvar,
			ident::local, level, type));
	}
	else
	{
		try
		{
			counter = (variable*) findid(cntvar);
			if (counter->kind != ident::var)
				throw "variable expected";
		}
		catch (out_of_range)
		{
			if (explic)
				throw "undeclared identifier '" + cntvar + "'";
			out << "float ";
			me.idents.push_back(counter = new variable(cntvar,
				ident::local, level, findtype("SINGLE")));
		}
	}
	out << counter->cname() << " = ";
	if (lex.token != "=")
		throw "assignment expected";
	lex.next();
	parse_expr();
	out << ";" << newline() << "float __limit = ";
	if (!(lex.token ^ "TO"))
		throw "TO expected";
	lex.next();
	parse_expr();
	if (lex.token ^ "STEP")
	{
		lex.next();
		if (lex.token == "-" || lex.token == "+")
		{
			if (lex.token == "-")
				negative = true;
			step = lex.token;
			lex.next();
		}
		else
			step = "";
		if (lex.tokkind != lexer::num)
			throw "number expected";
		step += lex.token;
		lex.next();
	}
	else if (lex.tokkind != lexer::nl)
		throw "STEP or end of statement expected";
	out << ";" << newline() << "while ( " << counter->cname()
		<< (negative ? " >=" : " <=") << " __limit )"
		<< newline() << "{";
	indent++;
	lex.next();
	parse_block();
	if (!(lex.token ^ "NEXT"))
		throw "NEXT expected";
	lex.next();
	if (lex.tokkind != lexer::nl)
		lex.next();
	out << counter->cname() << " += " << step << ";";
	indent--; loops--;
	out << newline() << "}";
	indent--;
	out << newline() << "}";
	leave();
}


void parser::parse_foreach()
{
	lex.next();
	if (lex.tokkind != lexer::id)
		throw "variable expected";
	variable *counter, *index = NULL;
	out << "// for-each" << newline() << "{" ;
	enter();
	indent++; loops++;
	out << newline();
	try
	{
		string cntvar = lex.token;
		lex.next();
		if (lex.token ^ "AS")
		{
			lex.next();
			udt* type = parse_typename();
			if (type->kind == udt::struc)
				throw "illegal counter type";
			me.idents.push_back(counter = new variable(cntvar,
				ident::local, level, type));
			if (type->kind == udt::cls)
				out << "bpp::ref<class ";
			out << type->cname();
			if (type->kind == udt::cls)
				out << "> ";
			out << " ";
			out << counter->cname() << "; ";
		}
		else
		{
			counter = (variable*) findid(cntvar);
			if (counter->kind != ident::var)
				throw "variable expected";
		}
		if (lex.token == ",")
		{
			lex.next();
			index = counter;
			counter = (variable*) findid(lex.token);
			if (counter->kind != ident::var)
				throw "variable expected";
			if (index->type != findtype("STRING"))
				throw "type mismatch";
			lex.next();
		}
	}
	catch (out_of_range)
	{
		throw "undeclared identifier '" + lex.token + "'";
	}
	if (lex.token != "IN")
		throw "IN expected";
	lex.next();
	out << "bpp::iterator* __iter = ";
	udt* type = parse_expr(false, false, false, true);
	if (type->kind != ident::cls || counter->type->kind != ident::cls)
		if (type != counter->type)
			throw "type mismatch";
	out << ".iterate();" << newline();
	if (lex.tokkind != lexer::nl)
		throw "STEP or end of statement expected";
	out << "while (__iter->next(";
	if (index)
		out << index->cname() << ", ";
	out << "bpp::ptr(" << counter->cname() << ")))" << newline() << "{";
	indent++;
	lex.next();
	parse_block();
	if (!(lex.token ^ "NEXT"))
		throw "NEXT expected";
	lex.next();
	if (lex.tokkind != lexer::nl)
		lex.next();
	indent--; loops--;
	out << newline() << "}";
	out << newline() << "delete __iter;";
	indent--;
	out << newline() << "}";
	leave();
}


void parser::parse_print()
{
	out << "print";
	while (true)
	{
		out << " << ";
		lex.next();
		parse_expr();
		if (!(lex.token ^ ","))
			break;
	}
	if (lex.token ^ ";")
		lex.next();
	else
		out << " << string(\"\\r\\n\")";
	out << "; ";
}


void parser::parse_input()
{
	out << "input >> ";
	lex.next();
	parse_expr(true, true);
	out << "; ";
}


void parser::parse_lineinput()
{
	lex.next();
	if (!(lex.token ^ "INPUT"))
		throw "INPUT expected";
	lex.next();
	out << "input.line(";
	if (!(parse_expr(true, true)->name ^ "STRING"))
		throw "STRING lvalue expected";
	out << ");";
}


void parser::parse_redim()
{
	lex.next();
	if (lex.tokkind != lexer::id)
		throw "identifier expected";
	variable* var;
	try
	{
		var = (variable*)findid(lex.token);
	}
	catch (out_of_range)
	{
		throw "undeclared array '" + lex.token + "'";
	}
	if (var->kind != ident::var)
		throw "array expected";
	if (!var->array)
		throw "array expected";
	out << var->cname() << ".redim(";
	lex.next();
	if (lex.token != "(")
		throw "brace expected";
	lex.next();
	int i = 0;
	while (lex.token != ")")
	{
		if (++i > 4)
			throw "too many dimensions";
		parse_expr();
		if (lex.token ^ "TO")
		{
			out << ", ";
			lex.next();
			parse_expr();
		}
		else
			out << ", 0";
		if (lex.token == ",")
		{
			out << ", ";
			lex.next();
		}
		else if (lex.token != ")")
			throw "closing brace expected";
		i++;
	}
	out << ");";
	lex.next();
	if (lex.token ^ "AS")
	{
		lex.next();
		parse_typename();
	}
}


void parser::write_dll_loader()
{
	out << newline() << "// DLL loader" << newline() << "{";
	indent++; out << newline();
	vector<string>::iterator i;
	for (i = dlls.begin(); i != dlls.end(); i++)
		out << "bpp::loadlib(" << (*i) << ");" << newline();
	indent--;
	out << newline() << "}" << newline();
}


void parser::write_dll_unloader()
{
	out << newline() << "/* DLL unloader */ {"
		<< " bpp::unloadlibs(); }" << newline();
}

void parser::parse_initialization()
{
	lex.next();
	if (lex.tokkind != lexer::nl)
		throw "end of statement expected";
	out << "struct initialization { initialization(); };" << newline();
	out << "initialization::initialization()" << newline() << "{";
	level++;
	indent++;
	// debug build
	if (devmode)
		out << newline() << "bpp::dbg_func(\"" << me.name << ".INITIALIZATION\");";
	lex.next();
	write_dll_loader();
	parse_block();
	// debug build
	if (devmode)
		out << newline() << "bpp::dbg_endfunc();";
	indent--;
	level--;
	if (!(lex.token ^ "INITIALIZATION"))
		throw "INITIALIZATION expected";
	out << newline() << "}";
	me.init = true;
	lex.next();
}


void parser::parse_finalization()
{
	lex.next();
	if (lex.tokkind != lexer::nl)
		throw "end of statement expected";
	out << "struct finalization { ~finalization(); };" << newline();
	out << "finalization::~finalization()" << newline() << "{";
	indent++;
	level++;
	// debug build
	if (devmode)
		out << newline() << "bpp::dbg_func(\"" << me.name << ".INITIALIZATION\");";
	lex.next();
	write_dll_unloader();
	parse_block();
	// debug build
	if (devmode)
		out << newline() << "bpp::dbg_endfunc();";
	level--;
	if (!(lex.token ^ "FINALIZATION"))
		throw "FINALIZATION expected";
	indent--;
	out << newline() << "}";
	me.final = true;
	lex.next();
}


void parser::parse_class(int scope)
{
	lex.next();
	if (lex.tokkind != lexer::id)
		throw "identifier expected";
	clas* cls;
	try
	{
		cls = (clas*)me.types[lex.token];
		if (cls->kind != ident::cls || cls->resolved)
			throw "duplicate identifier '" + cls->name + "'";
		cls->scope = scope;
	}
	catch (out_of_range)
	{
		cls = new clas(lex.token, scope);
		validate(cls->name);
		me.types.push_back(cls);
	}
	cls->resolved = true;
	out << "class " << cls->cname();
	lex.next();
	if (!(cls->name ^ "OBJECT"))
	{
		if (!(lex.token ^ "EXTENDS"))
			throw "EXTENDS expected";
		lex.next();
		cls->super = (clas*)parse_typename();
		if (cls->super->kind != udt::cls)
			throw "'" + cls->super->name + "' is not a class";
		out << ": public " << cls->super->cname();
	}
	out << newline() << "{";
	out << newline() << "public: " << cls->cname() << "(bool init = true): "
		<< cls->super->cname() << "(false) { if (init) { __refcnt++; Initialize(); __refcnt--; } }";
	out << newline() << "virtual bool __isbase(const type_info& ti)"
		<< "{ return (typeid(" << cls->cname() << ") == ti ) || "
		<< cls->super->cname() << "::__isbase(ti); }";
	out << newline() << "virtual bpp::string ClassName() { return \""
		<< me.name << "." << cls->name << "\"; }";
	cls->methods.push_back(cls->super->method("ClassName"));
	while (true)
	{
		int scope;
		lex.next();
		if (lex.tokkind == lexer::nl)
			continue;
		else if (lex.token ^ "END")
		{
			lex.next();
			if (!(lex.token ^ "CLASS"))
				throw "CLASS expected";
			lex.next();
			break;
		}
		else if (lex.token ^ "PRIVATE")
			scope = ident::priv;
		else if (lex.token ^ "PROTECTED")
			scope = ident::prot;
		else if (lex.token ^ "PUBLIC")
			scope = ident::pub;
		else
			throw "access specifier expected";
		out << newline();
		lex.next();
		if (lex.token ^ "SUB" || lex.token ^ "FUNCTION")
			parse_func(scope, true, cls);
		else if (lex.token ^ "PROPERTY")
			parse_property(scope, cls);
		else
		{
			while (true)
			{
				variable* prop = new variable("", scope);
				vector<string> props;
				if (lex.token == "(")
				{
					lex.next();
					while (true)
					{
						if (lex.tokkind != lexer::id)
							throw "identifier expected";
						try
						{
							throw "duplicate identifier '" +
								cls->member(lex.token)->name + "'";
						}
						catch (out_of_range)
						{
						}
						validate(lex.token);
						props.push_back(lex.token);
						lex.next();
						if (lex.token != ",")
							break;
						lex.next();
					}
					if (lex.token != ")")
						throw "closing brace expected";
					lex.next();
				}
				else
				{
					if (lex.tokkind != lexer::id)
						throw "identifier expected";
					try
					{
						throw "duplicate identifier '" +
							cls->member(lex.token)->name + "'";
					}
					catch (out_of_range)
					{
						validate(lex.token);
						props.push_back(lex.token);
						lex.next();
					}
				}
				if (lex.token == "(")
				{
					prop->array = true;
					lex.next();
					if (lex.token != ")")
						throw "closing brace expected";
					lex.next();
				}
				if (lex.token ^ "AS")
				{
					lex.next();
					prop->type = parse_typename();
				}
				else
					prop->type = findtype(prop->deftype());
				if (cls->scope == ident::pub && prop->type->scope < scope)
					throw "cannot use private type in public declaration";
				if (lex.token ^ "COLLECTION")
				{
					if (prop->array)
						throw "array of collections is not allowed";
					prop->col = variable::coll;
					lex.next();
				}
				else if (lex.token ^ "DICTIONARY")
				{
					if (prop->array)
						throw "array of dictionaries is not allowed";
					prop->col = variable::dict;
					lex.next();
				}
				//cls->props.push_back(prop);
				if (prop->array)
					out << "bpp::array<";
				if (prop->col == variable::coll)
					out << "bpp::collection<";
				else if (prop->col == variable::dict)
					out << "bpp::dictionary<";
				if (prop->type->kind == udt::cls)
					out << "bpp::ref<class ";
				out << prop->type->cname();
				if (prop->type->kind == udt::cls)
					out << "> ";
				if (prop->col)
					out << "> ";
				if (prop->array)
					out << ">";
				out << " ";
				vector<string>::iterator i;
				for (i = props.begin(); i != props.end(); i++)
				{
					variable* p = new variable(*i, prop->scope, prop->level, prop->type);
					p->array = prop->array;
					p->scope = prop->scope;
					p->col = prop->col;
					cls->props.push_back(p);
					if (i != props.begin())
						out << ", ";
					out << p->cname(false);
				}
				out << "; ";
				if (lex.token != ",")
					break;
				lex.next();
			}
		}
	}
	out << newline() << "};" << newline();
	out << "struct __rtti_" << cls->cname(false) << ": bpp::rtti" << newline() << "{" << newline();
	out << "\t__rtti_" << cls->cname(false) << "() { bpp::classlist::add(this); }" << newline();
	out << "\tbpp::ref<bpp::object> create() { return new class "
		<< cls->cname() << "; }" << newline();
	out << "\tstring name() { return \"" << me.name << "." << cls->name << "\"; }" << newline();
	out << "} __rtti_" << cls->cname(false) << ";" << newline();
}


void parser::parse_classdecl(int scope)
{
	lex.next();
	if (lex.tokkind != lexer::id)
		throw "identifier expected";
	clas* cls;
	try
	{
		cls = (clas*)me.types[lex.token];
		throw "duplicate identifier '" + cls->name + "'";
	}
	catch (out_of_range)
	{
		cls = new clas(lex.token, scope);
		validate(cls->name);
		cls->resolved = false;
		me.types.push_back(cls);
		out << "class " << cls->cname() << ";" << newline();
	}
	lex.next();
}

	
void parser::parse_raise()
{
	lex.next();
	if (!(parse_expr()->name ^ "EVENT"))
		throw "event expected";
	if (lex.token != "(")
		throw "braces expected";
	out << ".raise(";
	lex.next();
	parse_expr();
	if (lex.token != ")")
		throw "closing brace expected";
	out << ");";
	lex.next();
}


void parser::parse_try()
{
	// debug build
	if (devmode)
		out << "{ int __unwind_pos = bpp::dbg_savefunc();" << newline();
	out << "try" << newline() << "{";
	enter();
	indent++;
	lex.next();
	parse_block();
	indent--;
	leave();
	out << newline() << "}";
	if (lex.token ^ "TRY")
	{
		out << newline();
		// debug build
		if (devmode)
			out << "catch (...) { bpp::dbg_unwind(__unwind_pos); }";
		else
			out << "catch (...) { }";
	}
	else
	{
		set<udt*> handlers;
		bool done = false;
		while (!done)
		{
			if (lex.token ^ "TRY")
				break;
			else if (!(lex.token ^ "CATCH"))
				throw "CATCH expected";
			lex.next();
			enter();
			out << newline() << "catch (";
			if (lex.tokkind == lexer::nl)
			{
				out << "...";
				done = true;
			}
			else if (lex.tokkind == lexer::id)
			{
				variable* var = new variable(lex.token, ident::local, level);
				lex.next();
				if (lex.token ^ "AS")
				{
					lex.next();
					var->type = parse_typename();
				}
				else
					var->type = findtype(var->deftype());
				out << var->type->cname() << " " << var->cname();
				me.idents.push_back(var);
				if (handlers.find(var->type) != handlers.end())
					throw "duplicate exception handler";
				handlers.insert(var->type);
				if (lex.tokkind != lexer::nl)
					throw "end of statement expected";
			}
			else
				throw "identifier expected";
			out << ")" << newline() << "{";
			indent++;
			// debug build
			if (devmode)
				out << "bpp::dbg_unwind(__unwind_pos);" << newline();
			lex.next();
			parse_block();
			indent--;
			out << newline() << "}";
			leave();
		}
	}
	// debug build
	if (devmode)
		out << newline() << "}";
	lex.next();
}


void parser::parse_throw()
{
	out << "throw ";
	lex.next();
	parse_expr();
	out << ";";
}


void parser::parse_select()
{
	lex.next();
	if (!(lex.token ^ "CASE"))
		throw "CASE expected";
	lex.next();
	out << "{";
	indent++;
	out << newline() << "variant is = ";
	parse_expr();
	if (lex.tokkind != lexer::nl)
		throw "end of statement expected";
	out << ";" << newline() << "if (false);";
	while (lex.tokkind == lexer::nl)
		lex.next();
	bool done = false;
	while (!done)
	{
		if (lex.token ^ "SELECT")
			break;
		else if (!(lex.token ^ "CASE"))
			throw "CASE expected";
		out << newline() << "else";
		lex.next();
		if (lex.token ^ "ELSE")
		{
			done = true;
			lex.next();
		}
		else
		{
			out << " if (";
			while (true)
			{
				out << "(";
				parse_expr();
				if (lex.token ^ "TO")
				{
					out << " <= is && is <= ";
					lex.next();
					parse_expr();
				}
				else
					out << " == is";
				out << ")";
				if (lex.token != ",")
					break;
				out << " || ";
				lex.next();
			}
			out << ")";
		}
		if (lex.tokkind != lexer::nl)
			throw "end of statement expected";
		enter();
		out << newline() << "{";
		indent++;
		lex.next();
		parse_block();
		indent--;
		out << newline() << "}";
		leave();
	}
	indent--;
	out << newline() << "}";
	lex.next();
}


void parser::parse_builtin(int scope)
{
	lex.next();
	if (lex.token ^ "TYPE")
	{
		lex.next();
		if (lex.tokkind != lexer::id)
			throw "identifier expected";
		me.types.push_back(new structure(lex.token, scope));
		out << "// builtin type " << lex.token;
	}
	else if (lex.token ^ "CLASS")
	{
		lex.next();
		if (lex.tokkind != lexer::id)
			throw "identifier expected";
		clas* obj = new clas(lex.token, ident::pub);
		obj->methods.push_back(new function("Initialize", ident::prot,
			true, true, NULL));
		obj->methods.push_back(new function("Terminate", ident::prot,
			true, true, NULL));
		obj->methods.push_back(new function("ClassName", ident::pub,
			true, true, findtype("STRING")));
		{
			function* func;
			variable* var;
			var = new variable("Stream", ident::local, 1, obj);
			func = new function("SaveToStream", ident::pub, true, true, NULL);
			func->args.push_back(var);
			obj->methods.push_back(func);
			var = new variable("Stream", ident::local, 1, obj);
			func = new function("LoadFromStream", ident::pub, true, true, NULL);
			func->args.push_back(var);
			obj->methods.push_back(func);
		}
		me.types.push_back(obj);
		out << "// builtin class " << lex.token;
	}
	else
	{
		if (lex.tokkind != lexer::id)
			throw "identifier expected";
		me.types.push_back(new udt(lex.token));
		out << "// builtin " << lex.token;
	}
	lex.next();
}


void parser::parse_using()
{
	while (true)
	{
		lex.next();
		if (lex.tokkind != lexer::id)
			throw "identifier expected";
		lex.token = proper(lex.token);
		modules.push_back(new module(lex.token, *this, lib));
//		out << newline();
		lex.next();
		if (lex.token != ",")
			break;
	}
}


void parser::parse_alias()
{
	lex.next();
	if (lex.tokkind != lexer::id)
		throw "identifier expected";
	try
	{
		throw "duplicate identifier '" + me.idents[lex.token]->name + "'";
	}
	catch (out_of_range)
	{
	}
	string name = lex.token;
	lex.next();
	if (lex.token != "=")
		throw "assignment expected";
	lex.next();
	bool typealias = false;
	if (lex.token ^ "TYPE")
	{
		typealias = true;
		lex.next();
	}
	if (lex.tokkind != lexer::id)
		throw "identifier expected";
	try
	{
		if (typealias)
			me.types.push_back((udt*)new alias(name, me.types[lex.token], true));
		else
			me.idents.push_back(new alias(name, me.idents[lex.token]));
	}
	catch (out_of_range)
	{
		if (typealias)
			throw "typename expected";
		else
			throw "identifier expected";
	}
	lex.next();
}


void parser::parse_incr()
{
	lex.next();
	try
	{
		parse_expr(true, true);
	}
	catch (invalid_argument)
	{
		throw "lvalue expected";
	}
	if (lex.token == ",")
	{
		out << " += ";
		lex.next();
		parse_expr();
	}
	else
		out << "++";
	out << ";";
}


void parser::parse_decr()
{
	lex.next();
	try
	{
		parse_expr(true, true);
	}
	catch (invalid_argument)
	{
		throw "lvalue expected";
	}	
	if (lex.token == ",")
	{
		out << " -= ";
		lex.next();
		parse_expr();
	}
	else
		out << "--";
	out << ";";
}


void parser::parse_option()
{
	lex.next();
	if (lex.token ^ "EXPLICIT")
		explic = true;
	else if (lex.token ^ "CONSOLE")
	{
		lex.next();
		if (lex.token ^ "ON" || lex.tokkind == lexer::nl)
		{
			console = true;
			dll = false;
		}
		else if (lex.token ^ "OFF")
			console = false;
		else
			throw "ON or OFF expected";
	}
	else if (lex.token ^ "DLL")
	{
		lex.next();
		if (lex.token ^ "ON" || lex.tokkind == lexer::nl)
		{
			console = false;
			dll = true;
		}
		else if (lex.token ^ "OFF")
			dll = false;
		else
			throw "ON or OFF expected";
	}
	else if (lex.token ^ "APISTRING")
	{
		lex.next();
		if (lex.token ^ "ON" || lex.tokkind == lexer::nl)
			apistr = true;
		else if (lex.token ^ "OFF")
			apistr = false;
		else
			throw "ON or OFF expected";
	}
	else if (lex.token ^ "BYREF" || lex.token ^ "VBDLL")
		vbdll = true;
	else if (lex.token ^ "BYVAL")
		vbdll = false;
	else if (lex.token ^ "RESOURCE")
	{
		lex.next();
		if (lex.tokkind != lexer::str)
			throw "resource file expected";
		resfile = lex.token.substr(1, lex.token.length() - 2);
		if (!fexist(resfile))
			throw "resource file not found";
	}
	else
		throw "unknown option";
	lex.next();
}


void parser::parse_label()
{
	if (labels.find(lex.token) != labels.end())
		if (labels[lex.token])
			throw "duplicate label name";
	labels[lex.token] = true;
	out << "__label_" << lex.token << ":;";
	lex.next();
}


void parser::parse_goto()
{
	lex.next();
	if (lex.tokkind != lexer::id)
		throw "label expected";
	if (labels.find(lex.token) == labels.end())
		labels[lex.token] = false;
	out << "goto __label_" << lex.token << ";";
	lex.next();
}


void parser::parse_enum(int scope)
{
	lex.next();
	if (lex.tokkind == lexer::id)
	{
		validate(lex.token);
		alias* als = new alias(lex.token, findtype("LONG"), true);
		als->scope = scope;
		me.types.push_back((udt*)als);
		lex.next();
	}	
	if (lex.tokkind != lexer::nl)
		throw "newline expected";
	lex.next();
	int value = -1;
	while (true)
	{
		if (lex.token ^ "END")
			break;
		if (lex.tokkind != lexer::id)
			throw "identifier expected";
		constant* cons = new constant(lex.token, scope);
		validate(cons->name);
		lex.next();
		if (lex.token == "=")
		{
			lex.next();
			value = atof(parse_conexpr(cons->str).c_str());
			if (cons->str)
				throw "integer constant expected";
			cons->value = itos(value);	
		}
		else
			cons->value = itos(++value);
		cons->str = false;
		me.idents.push_back(cons);
		if (lex.tokkind == lexer::nl || lex.token == ",")
		{
			do
			lex.next();
			while (lex.tokkind == lexer::nl);
		}
		else
			throw "comma or newline expected";
	}
	lex.next();
	if (!(lex.token ^ "ENUM"))
		throw "ENUM expected";
	lex.next();
}
